import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import library
from leapp.libraries.common.testutils import produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import StorageInfo, FstabEntry, MountEntry, SystemdMountEntry, XFSPresence


class run_mocked(object):
    def __init__(self):
        self.called = 0
        self.args = None

    def __call__(self, args, split=True):
        self.called += 1
        self.args = args

        with_ftype = {'stdout': [
            "meta-data=/dev/loop0             isize=512    agcount=4, agsize=131072 blks",
            "         =                       sectsz=512   attr=2, projid32bit=1",
            "         =                       crc=1        finobt=0 spinodes=0",
            "data     =                       bsize=4096   blocks=524288, imaxpct=25",
            "         =                       sunit=0      swidth=0 blks",
            "naming   =version 2              bsize=4096   ascii-ci=0 ftype=1",
            "log      =internal               bsize=4096   blocks=2560, version=2",
            "         =                       sectsz=512   sunit=0 blks, lazy-count=1",
            "realtime =none                   extsz=4096   blocks=0, rtextents=0"]}

        without_ftype = {'stdout': [
            "meta-data=/dev/loop0             isize=512    agcount=4, agsize=131072 blks",
            "         =                       sectsz=512   attr=2, projid32bit=1",
            "         =                       crc=1        finobt=0 spinodes=0",
            "data     =                       bsize=4096   blocks=524288, imaxpct=25",
            "         =                       sunit=0      swidth=0 blks",
            "naming   =version 2              bsize=4096   ascii-ci=0 ftype=0",
            "log      =internal               bsize=4096   blocks=2560, version=2",
            "         =                       sectsz=512   sunit=0 blks, lazy-count=1",
            "realtime =none                   extsz=4096   blocks=0, rtextents=0"]}

        if "/var" in self.args:
            return without_ftype

        return with_ftype


def test_check_xfs_fstab(monkeypatch):
    fstab_data_no_xfs = {
        "fs_spec": "/dev/mapper/fedora-home",
        "fs_file": "/home",
        "fs_vfstype": "ext4",
        "fs_mntops": "defaults,x-systemd.device-timeout=0",
        "fs_freq": "1",
        "fs_passno": "2"}

    mountpoints = library.check_xfs_fstab([FstabEntry(**fstab_data_no_xfs)])
    assert not mountpoints

    fstab_data_xfs = {
        "fs_spec": "/dev/mapper/rhel-root",
        "fs_file": "/",
        "fs_vfstype": "xfs",
        "fs_mntops": "defaults",
        "fs_freq": "0",
        "fs_passno": "0"}

    mountpoints = library.check_xfs_fstab([FstabEntry(**fstab_data_xfs)])
    assert mountpoints == {"/"}


def test_check_xfs_mount(monkeypatch):
    mount_data_no_xfs = {
        "name": "tmpfs",
        "mount": "/run/snapd/ns",
        "tp": "tmpfs",
        "options": "rw,nosuid,nodev,seclabel,mode=755"}

    mountpoints = library.check_xfs_mount([MountEntry(**mount_data_no_xfs)])
    assert not mountpoints

    mount_data_xfs = {
        "name": "/dev/vda1",
        "mount": "/boot",
        "tp": "xfs",
        "options": "rw,relatime,seclabel,attr2,inode64,noquota"}

    mountpoints = library.check_xfs_mount([MountEntry(**mount_data_xfs)])
    assert mountpoints == {"/boot"}


def test_check_xfs_systemdmount(monkeypatch):
    systemdmount_data_no_xfs = {
        "node": "/dev/sda1",
        "path": "pci-0000:00:17.0-ata-2",
        "model": "TOSHIBA_THNSNJ512GDNU_A",
        "wwn": "0x500080d9108e8753",
        "fs_type": "ext4",
        "label": "n/a",
        "uuid": "5675d309-eff7-4eb1-9c27-58bc5880ec72"}

    mountpoints = library.check_xfs_systemdmount([SystemdMountEntry(**systemdmount_data_no_xfs)])
    assert not mountpoints

    systemdmount_data_xfs = {
        "node": "/dev/sda1",
        "path": "/var",
        "model": "n/a",
        "wwn": "n/a",
        "fs_type": "xfs",
        "label": "n/a",
        "uuid": "n/a"}

    mountpoints = library.check_xfs_systemdmount([SystemdMountEntry(**systemdmount_data_xfs)])
    assert mountpoints == {"/var"}


def test_is_xfs_without_ftype(monkeypatch):
    monkeypatch.setattr(library, "run", run_mocked())

    assert library.is_xfs_without_ftype("/var")
    assert ' '.join(library.run.args) == "/usr/sbin/xfs_info /var"

    assert not library.is_xfs_without_ftype("/boot")
    assert ' '.join(library.run.args) == "/usr/sbin/xfs_info /boot"


def test_check_xfs(monkeypatch):
    monkeypatch.setattr(library, "run", run_mocked())

    def consume_no_xfs_message_mocked(*models):
        yield StorageInfo()
    monkeypatch.setattr(api, "consume", consume_no_xfs_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())

    library.check_xfs()
    assert api.produce.called == 1
    assert len(api.produce.model_instances) == 1
    assert isinstance(api.produce.model_instances[0], XFSPresence)
    assert not api.produce.model_instances[0].present
    assert not api.produce.model_instances[0].without_ftype

    def consume_ignored_xfs_message_mocked(*models):
        mount_data = {
            "name": "/dev/vda1",
            "mount": "/boot",
            "tp": "xfs",
            "options": "rw,relatime,seclabel,attr2,inode64,noquota"}
        yield StorageInfo(mount=[MountEntry(**mount_data)])
    monkeypatch.setattr(api, "consume", consume_ignored_xfs_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())

    library.check_xfs()
    assert api.produce.called == 1
    assert len(api.produce.model_instances) == 1
    assert isinstance(api.produce.model_instances[0], XFSPresence)
    assert not api.produce.model_instances[0].present
    assert not api.produce.model_instances[0].without_ftype

    def consume_xfs_with_ftype_message_mocked(*models):
        fstab_data = {
            "fs_spec": "/dev/mapper/rhel-root",
            "fs_file": "/",
            "fs_vfstype": "xfs",
            "fs_mntops": "defaults",
            "fs_freq": "0",
            "fs_passno": "0"}
        yield StorageInfo(fstab=[FstabEntry(**fstab_data)])
    monkeypatch.setattr(api, "consume", consume_xfs_with_ftype_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())

    library.check_xfs()
    assert api.produce.called == 1
    assert len(api.produce.model_instances) == 1
    assert isinstance(api.produce.model_instances[0], XFSPresence)
    assert api.produce.model_instances[0].present
    assert not api.produce.model_instances[0].without_ftype

    def consume_xfs_without_ftype_message_mocked(*models):
        fstab_data = {
            "fs_spec": "/dev/mapper/rhel-root",
            "fs_file": "/var",
            "fs_vfstype": "xfs",
            "fs_mntops": "defaults",
            "fs_freq": "0",
            "fs_passno": "0"}
        yield StorageInfo(fstab=[FstabEntry(**fstab_data)])
    monkeypatch.setattr(api, "consume", consume_xfs_without_ftype_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())

    library.check_xfs()
    assert api.produce.called == 1
    assert len(api.produce.model_instances) == 1
    assert isinstance(api.produce.model_instances[0], XFSPresence)
    assert api.produce.model_instances[0].present
    assert api.produce.model_instances[0].without_ftype

    def consume_no_message_mocked(*models):
        yield None
    monkeypatch.setattr(api, "consume", consume_no_message_mocked)
    monkeypatch.setattr(api, "produce", produce_mocked())

    with pytest.raises(StopActorExecutionError):
        library.check_xfs()



import os

from leapp.snactor.fixture import current_actor_context
from leapp.models import SelinuxRelabelDecision


def test_schedule_no_relabel(current_actor_context):
    current_actor_context.feed(SelinuxRelabelDecision(set_relabel=False))
    current_actor_context.run()
    assert not os.path.isfile('/.autorelabel')


def test_schedule_relabel(current_actor_context):
    current_actor_context.feed(SelinuxRelabelDecision(set_relabel=True))
    current_actor_context.run()
    assert os.path.isfile('/.autorelabel')

    # lets cleanup so we possibly not affect further testing
    os.unlink('/.autorelabel')



from leapp.actors import Actor
from leapp.models import Report, OpenSshConfig
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import api
from leapp.libraries.common.reporting import report_generic


class OpenSshProtocolCheck(Actor):
    """
    Protocol configuration option was removed.

    Check the value of Protocol in OpenSSH server config file
    and warn about its deprecation if it is set. This option was removed
    in RHEL 7.4, but it might still be hanging around.
    """

    name = 'open_ssh_protocol'
    consumes = (OpenSshConfig, )
    produces = (Report, )
    tags = (ChecksPhaseTag, IPUWorkflowTag, )

    def process(self):
        openssh_messages = self.consume(OpenSshConfig)
        config = next(openssh_messages)
        if list(openssh_messages):
            api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')

        if config.protocol:
            report_generic(
                title='OpenSSH configured with removed configuration Protocol',
                summary='OpenSSH is configured with removed configuration '
                        'option Protocol. If this used to be for enabling '
                        'SSHv1, this is no longer supported in RHEL 8. '
                        'Otherwise this option can be simply removed.',
                severity='low')



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class KernelCmdlineArg(Model):
    """
    Single kernel command line argument to include to RHEL-8 kernel cmdline
    """
    topic = SystemInfoTopic

    key = fields.String()
    value = fields.String()



from leapp.models import SourceRHSMInfo
from leapp.libraries.common import rhsm
from leapp.libraries.common.mounting import NotIsolatedActions
from leapp.libraries.stdlib import api


@rhsm.with_rhsm
def scan():
    info = SourceRHSMInfo()
    context = NotIsolatedActions(base_dir='/')
    rhsm.scan_rhsm_info(context, info)
    api.produce(info)



from leapp.actors import Actor
from leapp.libraries.common import dnfplugin
from leapp.models import FilteredRpmTransactionTasks, TargetUserSpaceInfo, UsedTargetRepositories, XFSPresence
from leapp.tags import DownloadPhaseTag, IPUWorkflowTag


class DnfPackageDownload(Actor):
    """
    Actor that invokes DNF to download the RPMs required for the upgrade transaction.

    This actor uses the rhel-upgrade plugin to perform the download of RPM for the transaction and performing the
    transaction test, that is something like a dry run trying to determine the success of the upgrade.
    """

    name = 'dnf_package_download'
    consumes = (UsedTargetRepositories, FilteredRpmTransactionTasks, TargetUserSpaceInfo, XFSPresence)
    produces = ()
    tags = (IPUWorkflowTag, DownloadPhaseTag)

    def process(self):
        presence = next(self.consume(XFSPresence), XFSPresence())
        xfs_present = presence.present and presence.without_ftype
        used_repos = self.consume(UsedTargetRepositories)
        tasks = next(self.consume(FilteredRpmTransactionTasks), FilteredRpmTransactionTasks())
        target_userspace_info = next(self.consume(TargetUserSpaceInfo), None)

        dnfplugin.perform_rpm_download(tasks=tasks, used_repos=used_repos, target_userspace_info=target_userspace_info,
                                       xfs_present=xfs_present)



import selinux

from leapp.libraries.actor.systemfacts import get_selinux_status
from leapp.models import SELinuxFacts

# FIXME: create valid tests...


def test_selinux_enabled_enforcing(monkeypatch):
    """
    Test case SELinux is enabled in enforcing mode
    """
    monkeypatch.setattr(selinux, 'is_selinux_mls_enabled', lambda: 1)
    monkeypatch.setattr(selinux, 'security_getenforce', lambda: 1)
    monkeypatch.setattr(selinux, 'selinux_getenforcemode', lambda: [0, 1])
    monkeypatch.setattr(selinux, 'is_selinux_enabled', lambda: 1)
    monkeypatch.setattr(selinux, 'selinux_getpolicytype', lambda: [0, 'targeted'])
    expected_data = {'policy': 'targeted',
                     'mls_enabled': True,
                     'enabled': True,
                     'runtime_mode': 'enforcing',
                     'static_mode': 'enforcing'}
    assert SELinuxFacts(**expected_data) == get_selinux_status()


def test_selinux_enabled_permissive(monkeypatch):
    """
    Test case SELinux is enabled in permissive mode
    """
    monkeypatch.setattr(selinux, 'is_selinux_mls_enabled', lambda: 1)
    monkeypatch.setattr(selinux, 'security_getenforce', lambda: 0)
    monkeypatch.setattr(selinux, 'selinux_getenforcemode', lambda: [0, 0])
    monkeypatch.setattr(selinux, 'is_selinux_enabled', lambda: 1)
    monkeypatch.setattr(selinux, 'selinux_getpolicytype', lambda: [0, 'targeted'])
    expected_data = {'policy': 'targeted',
                     'mls_enabled': True,
                     'enabled': True,
                     'runtime_mode': 'permissive',
                     'static_mode': 'permissive'}
    assert SELinuxFacts(**expected_data) == get_selinux_status()


def test_selinux_disabled(monkeypatch):
    """
    Test case SELinux is disabled
    """
    monkeypatch.setattr(selinux, 'is_selinux_mls_enabled', lambda: 0)
    monkeypatch.setattr(selinux, 'security_getenforce', lambda: 0)
    monkeypatch.setattr(selinux, 'selinux_getenforcemode', lambda: [0, 0])
    monkeypatch.setattr(selinux, 'is_selinux_enabled', lambda: 0)
    monkeypatch.setattr(selinux, 'selinux_getpolicytype', lambda: [0, 'targeted'])
    expected_data = {'policy': 'targeted',
                     'mls_enabled': False,
                     'enabled': False,
                     'runtime_mode': 'permissive',
                     'static_mode': 'permissive'}
    assert SELinuxFacts(**expected_data) == get_selinux_status()


class MockNoConfigFileOSError(object):
    def __init__(self):
        raise OSError


def test_selinux_disabled_no_config_file(monkeypatch):
    """
    Test case SELinux is disabled
    """
    monkeypatch.setattr(selinux, 'is_selinux_mls_enabled', lambda: 0)
    monkeypatch.setattr(selinux, 'security_getenforce', lambda: 0)
    monkeypatch.setattr(selinux, 'selinux_getenforcemode', MockNoConfigFileOSError)
    monkeypatch.setattr(selinux, 'is_selinux_enabled', lambda: 0)
    monkeypatch.setattr(selinux, 'selinux_getpolicytype', lambda: [0, 'targeted'])
    expected_data = {'policy': 'targeted',
                     'mls_enabled': False,
                     'enabled': False,
                     'runtime_mode': 'permissive',
                     'static_mode': 'disabled'}

    assert SELinuxFacts(**expected_data) == get_selinux_status()



from leapp.libraries.common import mounting, rhsm
from leapp.libraries.stdlib import api
from leapp.models import TargetRHSMInfo


def process():
    info = next(api.consume(TargetRHSMInfo), None)
    if info and info.release:
        rhsm.set_release(mounting.NotIsolatedActions(base_dir='/'), info.release)



from sos.plugins import Plugin, RedHatPlugin


class Leapp(Plugin, RedHatPlugin):
    """
    This plugin is used to gather all information necessary for reporting
    problems in regards to the leapp application.
    """

    plugin_name = 'leapp'
    packages = ('leapp', 'leapp-repository')

    def setup(self):
        self.add_copy_spec([
            '/var/lib/leapp/leapp.db',
            '/var/log/leapp/dnf-debugdata/',
            '/var/log/leapp/leapp-upgrade.log',
            '/var/log/leapp/leapp-report.txt',
            '/var/log/leapp/dnf-plugin-data.txt'
        ])



import operator

import six

from leapp.libraries.stdlib import api


OP_MAP = {
    '>': operator.gt,
    '>=': operator.ge,
    '<': operator.lt,
    '<=': operator.le
}


def _version_to_tuple(version):
    """Converts the version string ``major.minor`` to ``(major, minor)`` int tuple."""
    major, minor = version.split('.')
    return (int(major), int(minor))


def _validate_versions(versions):
    """Raise ``TypeError`` if provided versions are not strings in the form ``<integer>.<integer>``."""
    for ver in versions:
        split = ver.split('.')
        if not len(split) == 2 or not all(x.isdigit() for x in split):
            raise ValueError("Versions have to be in the form of '<integer>.<integer>' "
                             "but provided was '{}'".format(versions))


def _simple_versions(versions):
    """Return ``True`` if provided versions are list of strings without comparison operators."""
    return all(len(v.split()) == 1 for v in versions)


def _cmp_versions(versions):
    """Return ``True`` if provided versions are list of strings with comparison operators."""
    split = [v.split() for v in versions]
    if not all(len(s) == 2 for s in split):
        return False

    return all(s[0] in OP_MAP for s in split)


def matches_version(match_list, detected):
    """
    Check if the `detected` version meets the criteria specified in `match_list`.

    :param match_list: specification of versions to check against
    :type match_list: list or tuple of strings in one of the two following forms:
                      ``['>'|'>='|'<'|'<='] <integer>.<integer>`` form, where elements are ANDed,
                      meaning that ``['>= 7.6', '< 7.8']`` would match for ``'7.6'``, and ``'7,7'`` only.
                      ``<integer>.<integer>`` form, where elements are ORed, meaning that
                      ``['7.6', '7.7']`` would match for ``'7.6'``, and ``'7,7'`` only.
                      These two forms cannot be mixed, otherwise ``ValueError`` is raised.
    :param detected: detected version
    :type detected: string in the form ``<integer>.<integer>``
    :return: ``True`` if `detected` value matches one of the values in `match_list`, ``False`` otherwise
    :rtype: bool
    """
    if not isinstance(match_list, (list, tuple)):
        raise TypeError("Versions to check against have to be a list or tuple "
                        "but provided was {}: '{}'".format(type(match_list), match_list))
    if not all(isinstance(e, six.string_types) for e in match_list):
        raise TypeError("Versions to check against have to be a list or tuple of strings "
                        "but provided was {}: '{}'".format([type(e) for e in match_list], match_list))
    if not isinstance(detected, six.string_types):
        raise TypeError("Detected version has to be a string "
                        "but provided was {}: '{}'".format(type(detected), detected))
    _validate_versions([detected])

    if _simple_versions(match_list):
        # match_list = ['7.6', '7.7', '7.8', '7.9']
        _validate_versions(match_list)
        return detected in match_list
    elif _cmp_versions(match_list):
        detected = _version_to_tuple(detected)
        # match_list = ['>= 7.6', '< 7.10']
        _validate_versions([s.split()[1] for s in match_list])
        for match in match_list:
            op, ver = match.split()
            ver = _version_to_tuple(ver)
            if not OP_MAP[op](detected, ver):
                return False
        return True
    else:
        raise ValueError("Versions have to be a list or tuple of strings in the form "
                         "'['>'|'>='|'<'|'<='] <integer>.<integer>' or "
                         "'<integer>.<integer>' but provided was '{}'".format(match_list))


def matches_source_version(*match_list):
    """
    Check if one of provided source versions matches the configured one.

    :param match_list: specification of versions to check against
    :type match_list: strings, for details see argument ``match_list`` of function :func:`matches_version`.
    """
    source_version = api.current_actor().configuration.version.source
    return matches_version(match_list, source_version)


def matches_target_version(*match_list):
    """
    Check if one of provided target versions matches the configured one.

    :param match_list: specification of versions to check against
    :type match_list: strings, for details see argument ``match_list`` of function :func:`matches_version`.
    """
    target_version = api.current_actor().configuration.version.target
    return matches_version(match_list, target_version)



import contextlib
import os
import shutil

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import mounting, utils
from leapp.libraries.stdlib import CalledProcessError, api, run


def _overlay_disk_size():
    """
    Convenient function to retrieve the overlay disk size
    """
    try:
        env_size = os.getenv('LEAPP_OVL_SIZE', default='2048')
        disk_size = int(env_size)
    except ValueError:
        disk_size = 2048
        api.current_logger().warn(
            'Invalid "LEAPP_OVL_SIZE" environment variable "%s". Setting default "%d" value', env_size, disk_size
        )
    return disk_size


def cleanup_scratch(scratch_dir, mounts_dir):
    """
    Function to cleanup the scratch directory
    """
    api.current_logger().debug('Cleaning up mounts')
    if os.path.ismount(mounts_dir):
        api.current_logger().debug('Mounts directory is a mounted disk image - Unmounting.')
        try:
            run(['/bin/umount', '-fl', mounts_dir])
            api.current_logger().debug('Unmounted mounted disk image.')
        except (OSError, CalledProcessError) as e:
            api.current_logger().warn('Failed to umount %s - message: %s', mounts_dir, str(e))
    api.current_logger().debug('Recursively removing scratch directory %s.', scratch_dir)
    shutil.rmtree(scratch_dir, onerror=utils.report_and_ignore_shutil_rmtree_error)
    api.current_logger().debug('Recursively removed scratch directory %s.', scratch_dir)


@utils.clean_guard(cleanup_function=cleanup_scratch)
def _create_mount_disk_image(scratch_dir, mounts_dir):
    """
    Creates the mount disk image, for cases when we hit XFS with ftype=0
    """
    diskimage_path = os.path.join(scratch_dir, 'diskimage')
    disk_size = _overlay_disk_size()

    api.current_logger().debug('Attempting to create disk image with size %d MiB at %s', disk_size, diskimage_path)
    utils.call_with_failure_hint(
        cmd=['/bin/dd', 'if=/dev/zero', 'of={}'.format(diskimage_path), 'bs=1M', 'count={}'.format(disk_size)],
        hint='Please ensure that there is enough diskspace in {} at least {} MiB are needed'.format(
            scratch_dir, disk_size)
    )

    api.current_logger().debug('Creating ext4 filesystem in disk image at %s', diskimage_path)
    try:
        utils.call_with_oserror_handled(cmd=['/sbin/mkfs.ext4', '-F', diskimage_path])
    except CalledProcessError as e:
        api.current_logger().error('Failed to create ext4 filesystem %s', exc_info=True)
        raise StopActorExecutionError(
            message=str(e)
        )

    return diskimage_path


def _create_mounts_dir(scratch_dir, mounts_dir):
    """
    Prepares directories for mounts
    """
    api.current_logger().debug('Creating mount directories.')
    try:
        utils.makedirs(mounts_dir)
        api.current_logger().debug('Done creating mount directories.')
    except OSError:
        api.current_logger().error('Failed to create mounting directories %s', mounts_dir, exc_info=True)

        # This is an attempt for giving the user a chance to resolve it on their own
        raise StopActorExecutionError(
            message='Failed to prepare environment for package download while creating directories.',
            details={
                'hint': 'Please ensure that {scratch_dir} is empty and modifiable.'.format(scratch_dir=scratch_dir)
            }
        )


@contextlib.contextmanager
def _mount_dnf_cache(overlay_target):
    """
    Convenience context manager to ensure bind mounted /var/cache/dnf and removal of the mount.
    """
    with mounting.BindMount(
            source='/var/cache/dnf',
            target=os.path.join(overlay_target, 'var', 'cache', 'dnf')) as cache_mount:
        yield cache_mount


@contextlib.contextmanager
def _prepare_mounts(mounts_dir, scratch_dir, cleanup, detach, xfs_present):
    """
    A context manager function that prepares the scratch userspace and creates a mounting target directory. In case of
    XFS with ftype=0 it will trigger the creation of the disk image and mounts it and ensure the cleanup after leaving
    the context.
    """
    if cleanup:
        cleanup_scratch(scratch_dir, mounts_dir)
    _create_mounts_dir(scratch_dir=scratch_dir, mounts_dir=mounts_dir)
    if xfs_present:
        mount = mounting.LoopMount(source=_create_mount_disk_image(scratch_dir, mounts_dir),
                                   target=mounts_dir, config=mounting.MountConfig.MountOnly)
    else:
        mount = mounting.NullMount(target=mounts_dir)
    with mount:
        yield mount


@contextlib.contextmanager
def create_source_overlay(mounts_dir, scratch_dir, xfs_present, cleanup=True, detach=False):
    """
    Context manager that prepares the source system overlay and yields the mount.
    """
    api.current_logger().debug('Creating source overlay in {scratch_dir} with mounts in {mounts_dir}'.format(
        scratch_dir=scratch_dir, mounts_dir=mounts_dir))
    with _prepare_mounts(mounts_dir, scratch_dir, cleanup, detach, xfs_present) as mounts:
        with mounting.OverlayMount(name='source_overlay', source='/', workdir=mounts.target) as overlay:
            with _mount_dnf_cache(overlay.target):
                yield overlay



#
# Functions for probing SCTP usage by DLM
#
import re

from six.moves.configparser import SafeConfigParser

from leapp.libraries.stdlib import api


def check_dlm_cfgfile():
    """Parse DLM config file"""
    fname = "/etc/dlm/dlm.conf"

    try:
        with open(fname, 'r') as fp:
            cfgs = '[dlm]\n' + fp.read()
    except (OSError, IOError):
        return False

    cfg = SafeConfigParser()
    try:
        cfg.read_string(cfgs)
    except AttributeError:
        # Python2 ConfigParser doesn't have cfg.read_string
        from cStringIO import StringIO
        cfg.readfp(StringIO(cfgs))

    if not cfg.has_option('dlm', 'protocol'):
        return False

    proto = cfg.get('dlm', 'protocol').lower()
    return proto in ['sctp', 'detect', '1', '2']


def check_dlm_sysconfig():
    """Parse /etc/sysconfig/dlm"""
    regex = re.compile('^[^#]*DLM_CONTROLD_OPTS.*=.*(?:--protocol|-r)[ =]*([^"\' ]+).*', re.IGNORECASE)

    try:
        with open('/etc/sysconfig/dlm', 'r') as fp:
            lines = fp.readlines()
    except (OSError, IOError):
        return False

    for line in lines:
        if regex.match(line):
            proto = regex.sub('\\1', line).lower().strip()
            if proto in ['sctp', 'detect']:
                return True

    return False


def is_dlm_using_sctp():
    if check_dlm_cfgfile():
        api.current_logger().info('DLM is configured to use SCTP on dlm.conf.')
        return True

    if check_dlm_sysconfig():
        api.current_logger().info('DLM is configured to use SCTP on sysconfig.')
        return True

    return False



from leapp.snactor.fixture import current_actor_context
from leapp.models import StorageInfo, LsblkEntry
from leapp.reporting import Report


def test_actor_with_luks(current_actor_context):
    with_luks = [LsblkEntry(name='luks-132', maj_min='253:0', rm='0',
                            size='10G', ro='0', tp='crypt', mountpoint=''
                            )]

    current_actor_context.feed(StorageInfo(lsblk=with_luks))
    current_actor_context.run()
    assert current_actor_context.consume(Report)
    assert 'inhibitor' in current_actor_context.consume(Report)[0].flags


def test_actor_without_luks(current_actor_context):
    without_luks = [LsblkEntry(name='sda1', maj_min='8:0', rm='0',
                               size='10G', ro='0', tp='part', mountpoint='/boot'
                               )]

    current_actor_context.feed(StorageInfo(lsblk=without_luks))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



from leapp.actors import Actor
from leapp.libraries.actor import report
from leapp.reporting import Report
from leapp.tags import ReportPhaseTag, IPUWorkflowTag


class VerifyCheckResults(Actor):
    """
    Check all generated results messages and notify user about them.

    A report file containing all messages will be generated, together with log messages displayed
    to the user.
    """

    name = 'verify_check_results'
    consumes = (Report,)
    produces = ()
    tags = (ReportPhaseTag, IPUWorkflowTag)

    def process(self):
        results = list(self.consume(Report))
        inhibitors = [msg for msg in results if 'inhibitor' in msg.flags]
        high_sev_msgs = [msg for msg in results if msg.severity == 'high' and 'inhibitor' not in msg.flags]
        msgs_to_report = inhibitors + high_sev_msgs

        report_file = '/var/log/leapp/leapp-report.txt'
        error = report.generate_report(msgs_to_report, report_file)
        if error:
            self.report_error('Report Error: ' + error)
        else:
            self.log.info('Generated report at ' + report_file)

        if inhibitors:
            for e in inhibitors:
                self.report_error(e.title)

            self.report_error('Ending process due to errors found during checks, see {} for detailed report.'
                              .format(report_file))



from leapp.snactor.fixture import current_actor_context
from leapp.models import RpmTransactionTasks


def test_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(RpmTransactionTasks)
    assert current_actor_context.consume(RpmTransactionTasks)[0].local_rpms
    assert not current_actor_context.consume(RpmTransactionTasks)[0].to_install
    assert not current_actor_context.consume(RpmTransactionTasks)[0].to_remove
    assert not current_actor_context.consume(RpmTransactionTasks)[0].to_keep



from collections import namedtuple

import pytest

from leapp.libraries.common.config import version
from leapp.libraries.stdlib import api


class CurrentActorMocked(object):
    version = namedtuple('configuration', ['source', 'target'])('7.6', '8.0')
    configuration = namedtuple('configuration', ['version'])(version)


def test_version_to_tuple():
    assert version._version_to_tuple('7.6') == (7, 6)


def test_validate_versions():
    version._validate_versions(['7.6', '7.7'])
    with pytest.raises(ValueError):
        assert version._validate_versions(['7.6', 'z.z'])


def test_simple_versions():
    assert version._simple_versions(['7.6', '7.7'])
    assert not version._simple_versions(['7.6', '< 7.7'])


def test_cmp_versions():
    assert version._cmp_versions(['>= 7.6', '< 7.7'])
    assert not version._cmp_versions(['>= 7.6', '& 7.7'])


def test_matches_version_wrong_args():
    with pytest.raises(TypeError):
        version.matches_version('>= 7.6', '7.7')
    with pytest.raises(TypeError):
        version.matches_version([7.6, 7.7], '7.7')
    with pytest.raises(TypeError):
        version.matches_version(['7.6', '7.7'], 7.7)
    with pytest.raises(ValueError):
        version.matches_version(['>= 7.6', '> 7.7'], 'x.y')
    with pytest.raises(ValueError):
        version.matches_version(['>= 7.6', '7.7'], '7.7')
    with pytest.raises(ValueError):
        version.matches_version(['>= 7.6', '& 7.7'], '7.7')


def test_matches_version_fail():
    assert not version.matches_version(['> 7.6', '< 7.7'], '7.6')
    assert not version.matches_version(['> 7.6', '< 7.7'], '7.7')
    assert not version.matches_version(['> 7.6', '< 7.10'], '7.6')
    assert not version.matches_version(['> 7.6', '< 7.10'], '7.10')
    assert not version.matches_version(['7.6', '7.7', '7.10'], '7.8')


def test_matches_version_pass():
    assert version.matches_version(['7.6', '7.7', '7.10'], '7.7')
    assert version.matches_version(['> 7.6', '< 7.10'], '7.7')


def test_matches_source_version_pass(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked)
    assert version.matches_source_version('7.6', '7.7')


def test_matches_source_version_fail(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked)
    assert not version.matches_source_version('7.5', '7.7')


def test_matches_target_version_pass(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked)
    assert version.matches_target_version('8.0', '8.1')


def test_matches_source_targetn_fail(monkeypatch):
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked)
    assert not version.matches_target_version('8.2')



from leapp.actors import Actor
from leapp.libraries.actor import library
from leapp.models import OSReleaseFacts
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class OSReleaseCollector(Actor):
    """
    Provides data about System OS release.

    After collecting data from /etc/os-release file, a message with relevant data will be produced.
    """

    name = 'os_release_collector'
    consumes = ()
    produces = (Report, OSReleaseFacts,)
    tags = (IPUWorkflowTag, FactsPhaseTag,)

    def process(self):
        self.produce(library.get_os_release_info('/etc/os-release'))



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class OSReleaseFacts(Model):
    topic = SystemInfoTopic

    release_id = fields.String()
    name = fields.String()
    pretty_name = fields.String()
    version = fields.String()
    version_id = fields.String()
    variant = fields.Nullable(fields.String())  # isn't specified on some systems
    variant_id = fields.Nullable(fields.String())  # same as above



"""
tmeszaro@redhat.com

This script is called by the make test target.


There can be two arguments:

 1. ACTOR=myactor

 2. REPORT=myreport.xml

    Outputs xml report for the JUnit.


What is happening

 1. Checks cmd line arguments. There can be ACTOR and REPORT.

 2. Finds and registers all leapp repos in the BASE_REPO path.

 3. Checks if there are actor tests present.

 4. Runs pytest for each actor in BASE_REPO path.

 5. Combines reports from pytest runs if requested by --report argument
"""

import argparse
from glob import glob
import logging
import os
import shutil
import subprocess
import sys
import xml.etree.ElementTree as ET

from leapp.repository.scan import find_and_scan_repositories

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("run_pytest.py")

BASE_REPO = "repos"
REPORT_DIR = "reports/"


def snactor_register(path):
    """ Snactor registers all repos in @path.
    """
    cmd = "snactor repo find --path {PATH}".format(PATH=path)
    try:
        logger.info(" Registering leapp repositories. This may take a while.")
        return subprocess.check_output(cmd, shell=True)
    except OSError as exc:
        sys.stderr.write(str(exc) + '\n')
        return None


def combine_pytest_xmls(first, second):
    first.attrib.update({
        'errors': str(int(first.attrib['errors']) + int(second.attrib['errors'])),
        'failures': str(int(first.attrib['failures']) + int(second.attrib['failures'])),
        'skips': str(int(first.attrib['skips']) + int(second.attrib['skips'])),
        'tests': str(int(first.attrib['tests']) + int(second.attrib['tests'])),
        'time': str(float(first.attrib['time']) + float(second.attrib['time'])),
    })
    first.extend(second.findall('testcase'))


def produce_report(reports_dir, path):
    reports = (glob(reports_dir + '*.xml'))
    trees = [ET.parse(f) for f in reports]
    root = trees[0].getroot()
    for tree in trees[1:]:
        combine_pytest_xmls(root, tree.getroot())

    trees[0].write(path, encoding='utf-8', xml_declaration=True)


def test_libraries(libraries):
    for lib, name in libraries:
        os.environ['LEAPP_TESTED_LIBRARY'] = lib[0]
        cmd = pytest_cmd + [lib[0]]
        if args.report:
            cmd += ['--junit-xml={REPORT}'.format(REPORT=REPORT_DIR + name + '.xml')]
        logger.info(" Running pytest with: {PYTEST_CMD}".format(PYTEST_CMD=' '.join(cmd)))
        pytest_status.add(subprocess.call(cmd))


def test_actors(actors):
    for i, actor in enumerate(actors):
        # Run tests if actor has any.
        if not actor.tests:
            status = " Tests MISSING: {ACTOR} | class={CLASS}"
            status = status.format(ACTOR=actor.name, CLASS=actor.class_name)
            logger.critical(status)
        else:
            os.environ['LEAPP_TESTED_ACTOR'] = actor.full_path
            cmd = pytest_cmd + [actor.full_path]
            if args.report:
                cmd += ['--junit-xml={REPORT}'.format(REPORT=REPORT_DIR + actor.name + str(i) + '.xml')]
            logger.info(" Running pytest with: {PYTEST_CMD}".format(PYTEST_CMD=' '.join(cmd)))
            pytest_status.add(subprocess.call(cmd))


if __name__ == "__main__":
    pytest_cmd = ["pytest", "-v"]
    pytest_status = set()

    parser = argparse.ArgumentParser()
    parser.add_argument("--actor", help="name of the actor for which to run tests")
    parser.add_argument("--libraries", action='store_true', help="run tests for shared libraries")
    parser.add_argument("--report", help="filepath where to save report")
    args = parser.parse_args()

    shutil.rmtree(REPORT_DIR, ignore_errors=True)
    os.mkdir(REPORT_DIR)

    # Register repos.
    snactor_register(BASE_REPO)

    # Find and collect leapp repositories.
    repos = find_and_scan_repositories(BASE_REPO, include_locals=True)
    repos.load()

    libraries = [(r.libraries, r.name) for r in repos.repos if r.libraries]
    actors = repos.actors if not args.actor else (repos.lookup_actor(args.actor),)

    if args.actor and args.libraries:
        test_libraries(libraries)
        test_actors(actors)
    elif args.actor:
        test_actors(actors)
    elif args.libraries:
        test_libraries(libraries)
    else:
        test_libraries(libraries)
        test_actors(actors)

    if args.report:
        produce_report(REPORT_DIR, args.report)
    # Cleanup.
    shutil.rmtree(REPORT_DIR)
    sys.exit(max(pytest_status))



import os

from leapp.exceptions import StopActorExecution
from leapp.libraries.common import reporting
from leapp.libraries.stdlib import api
from leapp.libraries.stdlib.config import is_verbose
from leapp.models import InstalledUnsignedRPM


def skip_check():
    """ Check if has environment variable to skip this actor checks """
    if os.getenv('LEAPP_SKIP_CHECK_SIGNED_PACKAGES'):
        reporting.report_generic(title='Skipped signed packages check',
                                 severity='low',
                                 summary='Signed packages check skipped via LEAPP_SKIP_CHECK_SIGNED_PACKAGES env var')
        raise StopActorExecution()


def generate_report(packages):
    """ Generate a report if exists packages unsigned in the system """
    if not packages:
        return
    unsigned_packages_new_line = '\n'.join(['- ' + p for p in packages])
    unsigned_packages = ' '.join(packages)
    remediation = 'yum remove {}'.format(unsigned_packages)
    summary = 'The following packages have not been signed by Red Hat ' \
              'and may be removed in the upgrade process:\n{}'.format(unsigned_packages_new_line)
    reporting.report_with_remediation(
        title='Packages not signed by Red Hat found in the system',
        summary=summary,
        remediation=remediation,
        severity='high',
    )
    if is_verbose():
        api.show_message(summary)


def get_unsigned_packages():
    """ Get list of unsigned packages installed in the system """
    rpm_messages = api.consume(InstalledUnsignedRPM)
    data = next(rpm_messages, InstalledUnsignedRPM())
    if list(rpm_messages):
        api.current_logger().warning('Unexpectedly received more than one InstalledUnsignedRPM message.')
    unsigned_packages = set()
    unsigned_packages.update([pkg.name for pkg in data.items])
    unsigned_packages = list(unsigned_packages)
    unsigned_packages.sort()
    return unsigned_packages


def check_unsigned_packages():
    """ Check and generate reports if system contains unsigned installed packages"""
    generate_report(get_unsigned_packages())




from leapp.actors import Actor
from leapp.models import InitrdIncludes
from leapp.tags import IPUWorkflowTag, FinalizationPhaseTag
from leapp.libraries.stdlib import run, CalledProcessError
from leapp.exceptions import StopActorExecutionError


class InitrdInclude(Actor):
    """
    Regenerate RHEL-8 initrd and include files produced by other actors
    """

    name = 'initrdinclude'
    consumes = (InitrdIncludes, )
    produces = ()
    tags = (FinalizationPhaseTag, IPUWorkflowTag)

    def get_rhel8_kernel_version(self):
        kernels = run(["rpm", "-q", "kernel"], split=True)["stdout"]
        for kernel in kernels:
            version = kernel.split("-", 1)[1]
            if "el8" in version:
                return version
        raise StopActorExecutionError(
            "Cannot get version of the installed RHEL-8 kernel",
            details={"details": "\n".join(kernels)})

    def process(self):
        files = []

        for i in self.consume(InitrdIncludes):
            files += i.files

        if not files:
            self.log.debug("No additional files required to add into the initrd.")
            return

        # TODO(pstodulk): consume message that provides info about currently
        # # installed rpms; - it doesn't exist yet
        try:
            kernel_version = self.get_rhel8_kernel_version()
            self.log.debug("The detected RHEL 8 kernel: {}".format(kernel_version))
        except CalledProcessError as e:
            # just hypothetic check, it should not die
            # NOTE(pstodulk): raise an exception and stop leapp execution now?
            raise StopActorExecutionError(
                "Cannot get info about the installed kernel.",
                details={"details": str(e)})
        try:
            cmd = ["dracut", "--install", " ".join(files), "-f", "--kver", kernel_version]
            run(cmd)
        except CalledProcessError as e:
            # NOTE(pstodulk) same note as above
            raise StopActorExecutionError(
                "Cannot regenerate dracut image.",
                details={"details": str(e)})



import os


class produce_mocked(object):
    def __init__(self):
        self.called = 0
        self.model_instances = []

    def __call__(self, *model_instances):
        self.called += 1
        self.model_instances.append(model_instances[0])


class report_generic_mocked(object):
    def __init__(self):
        self.called = 0
        self.report_fields = None

    def __call__(self, **report_fields):
        self.called += 1
        self.report_fields = report_fields


def make_IOError(error):
    '''
    Create an IOError instance

    Create an IOError instance with the given error number.

    :param error: the error number, e.g. errno.ENOENT
    '''
    return IOError(error, os.strerror(error))


def make_OSError(error):
    '''
    Create an OSError instance

    Create an OSError instance with the given error number.

    :param error: the error number, e.g. errno.ENOENT
    '''
    return OSError(error, os.strerror(error))



from leapp.libraries import stdlib
from leapp.libraries.actor import scankernel
from leapp.libraries.stdlib import api

TARGET_KERNEL_VERSION = '1.2.3-4.el8.x86_64'
TARGET_KERNEL = 'kernel-{}'.format(TARGET_KERNEL_VERSION)
OLD_KERNEL = 'kernel-0.1.2-3.el7.x86_64'


def mocked_run_with_target_kernel(*args, **kwargs):
    return {'stdout': [TARGET_KERNEL, OLD_KERNEL]}


def mocked_run_without_target_kernel(*args, **kwargs):
    return {'stdout': [OLD_KERNEL]}


def test_scaninstalledkernel(monkeypatch):
    result = []
    monkeypatch.setattr(stdlib, 'run', mocked_run_with_target_kernel)
    monkeypatch.setattr(api, 'produce', result.append)
    scankernel.process()
    assert result and result[0].version == TARGET_KERNEL_VERSION


def test_scaninstalledkernel_missing(monkeypatch):
    result = []
    monkeypatch.setattr(stdlib, 'run', mocked_run_without_target_kernel)
    monkeypatch.setattr(api, 'produce', result.append)
    scankernel.process()
    assert not result



from leapp.actors import Actor
from leapp.models import Report, TcpWrappersFacts, InstalledRedHatSignedRPM
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import api
from leapp.libraries.actor.library import config_affects_daemons
from leapp.libraries.common.rpms import create_lookup
from leapp.libraries.common.reporting import report_with_links


DAEMONS = [
    ("audit", ["auditd"]),
    ("bacula", ["bacula"]),
    ("conserver", ["conserver"]),
    ("cyrus-imapd", ["imap", "pop3", "imaps", "pop3s", "sieve", "lmtp"]),
    ("dovecot", ["imap", "pop3", "imaps", "pop3s", "sieve", "lmtp"]),
    ("nfs-utils", ["mountd", "statd"]),
    ("openssh-server", ["sshd"]),
    ("proftpd", ["proftpd"]),
    ("quota", ["rquotad"]),
    ("rpcbind", ["rpcbind"]),
    # ("sendmail", ["sendmail"]),  # Handled in CheckSendmail Actor
    ("slapi-nis", ["slapi-nis"]),
    ("socat", ["socat"]),
    ("stunnel", ["stunnel"]),
    ("tftp-server", ["tftpd"]),
    # ("vsftpd", ["vsftpd"]),  # Handled in VsftpdConfigCheck Actor
    ("xinetd", ["xinetd"]),
]


class TcpWrappersCheck(Actor):
    """
    Check the list of packages previously compiled with TCP wrappers support
    and check whether they have some rules configured in
    /etc/hosts.{allow,deny} that are no longer honored by the daemons and
    might make the service more accessible than expected.
    """

    name = 'tcp_wrappers_check'
    consumes = (TcpWrappersFacts, InstalledRedHatSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        # Consume a single TCP Wrappers message
        tcp_wrappers_messages = self.consume(TcpWrappersFacts)
        tcp_wrappers_facts = next(tcp_wrappers_messages)
        if list(tcp_wrappers_messages):
            api.current_logger().warning('Unexpectedly received more than one TcpWrappersFacts message.')

        # Convert installed packages message to list
        packages = create_lookup(InstalledRedHatSignedRPM, field='items', key='name')

        found_packages = config_affects_daemons(tcp_wrappers_facts, packages, DAEMONS)

        if found_packages:
            report_with_links(title='TCP Wrappers configuration affects some installed packages',
                              summary=('tcp_wrappers support has been removed in RHEL-8. '
                                       'There is some configuration affecting installed packages (namely {}) '
                                       'in /etc/hosts.deny or /etc/hosts.allow, which '
                                       'is no longer going to be effective after update. '
                                       'Please migrate it manually.'.format(', '.join(found_packages))),
                              links=[{'title': 'Replacing TCP Wrappers in RHEL 8',
                                      'href': 'https://access.redhat.com/solutions/3906701'}],
                              severity='high',
                              flags=['inhibitor'])



import os
from collections import namedtuple

from leapp.libraries import stdlib
from leapp.libraries.stdlib import api
from leapp.models import InstalledTargetKernelVersion

KernelInfo = namedtuple('KernelInfo', ('kernel_path', 'initrd_path'))


def get_kernel_info(message):
    kernel_name = 'vmlinuz-{}'.format(message.version)
    initrd_name = 'initramfs-{}.img'.format(message.version)
    kernel_path = os.path.join('/boot', kernel_name)
    initrd_path = os.path.join('/boot', initrd_name)

    target_version_bootable = True
    if not os.path.exists(kernel_path):
        target_version_bootable = False
        api.current_logger().warning('Mandatory kernel %s does not exist', kernel_path)
    if not os.path.exists(initrd_path):
        target_version_bootable = False
        api.current_logger().warning('Mandatory initrd %s does not exist', initrd_path)

    if target_version_bootable:
        return KernelInfo(kernel_path=kernel_path, initrd_path=initrd_path)

    api.current_logger().warning('Skipping check due to missing mandatory files')
    return None


def update_default_kernel(kernel_info):
    try:
        stdlib.run(['grubby', '--info', kernel_info.kernel_path])
    except stdlib.CalledProcessError:
        api.current_logger().error('Expected kernel %s to be installed at the boot loader but cannot be found.',
                                   kernel_info.kernel_path)
    except OSError:
        api.current_logger().error('Could not check for kernel existence in boot loader. Is grubby installed?')
    else:
        try:
            stdlib.run(['grubby', '--set-default', kernel_info.kernel_path])
        except (OSError, stdlib.CalledProcessError):
            api.current_logger().error('Failed to set default kernel to: %s', kernel_info.kernel_path, exc_info=True)


def process():
    message = next(api.consume(InstalledTargetKernelVersion), None)
    if not message:
        api.current_logger().warning(('Skipped - Forcing checking and setting default boot entry to target kernel'
                                      ' version due to missing message'))
        return

    try:
        current_default_kernel = stdlib.run(['grubby', '--default-kernel'])['stdout'].strip()
    except (OSError, stdlib.CalledProcessError):
        api.current_logger().warning('Failed to query grubby for default kernel', exc_info=True)
        return

    kernel_info = get_kernel_info(message)
    if not kernel_info:
        return

    if current_default_kernel != kernel_info.kernel_path:
        api.current_logger().warning(('Current default boot entry not target kernel version: Current default: %s.'
                                      'Forcing default kernel to %s'),
                                     current_default_kernel, kernel_info.kernel_path)
        update_default_kernel(kernel_info)



from leapp.actors import Actor
from leapp.libraries.actor.library import check_xfs
from leapp.models import StorageInfo, XFSPresence
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckXFS(Actor):
    """
    Check if XFS filesystem is in use.

    If XFS filesystem without ftype is in use, produce a message to influence
    PrepareUpgradeTransaction actor about necessary steps during execution.
    """

    name = 'check_xfs'
    consumes = (StorageInfo,)
    produces = (XFSPresence,)
    tags = (ChecksPhaseTag, IPUWorkflowTag,)

    def process(self):
        check_xfs()



from leapp.reporting import report


"""
Detail requirements:
    summary : str

Example:
    {'summary': 'Some text here'}
"""
RenderGeneric = {
    'html': '<h2 class="report-title">{{ title }}</h2><p class="report-summary">{{ summary }}</p>',
    'plaintext': '{{ title }}\n{{ summary }}'
}


"""
Detail requirements:
    summary : str
    remediation : str

Example:
    {'summary': 'Here is a summary', 'remediation': 'and --here-is -a solution'}
"""
RenderWithRemediation = {
    'html': ('<h2 class="report-title">{{ title }}</h2>'
             '<p class="report-summary">{{ summary }}</p>'
             '<h3>Remediation:</h3>'
             '<pre class="report-code">{{ remediation }}</pre>'),
    'plaintext': '{{ title }}\n{{ summary }}\nRemediation:\n{{ remediation }}'
}


"""
Detail requirements:
    summary : str
    links : List[{href:str, title:str}]

Example:
    {'summary': 'Summary with links',
     'links': [{'href': 'https://some.important.link.com', 'title': 'Knowledge base article'}]}
"""
RenderWithLinks = {
    'html': RenderGeneric['html'] +  # noqa: E131,W504
             ('<ul class="report-list">{% for link in links %}'
              '<li class="report-list-item">'
              ' <a href="{{ link.href }}" target="_blank" class="report-link">{{ link.title }}</a>'
              '</li>'
              '{% endfor %}</ul>'),

    'plaintext': ('{{ title }}\n'
                  '{{ summary }}\n'
                  'Links:\n'
                  '{% for link in links %}{{ link.title }} - {{ link.href }}{% endfor %}')
}


def _validate_args(title, summary):
    if not title:
        raise ValueError('Title not provided in the report.')
    if not summary:
        raise ValueError('Summary not provided in the report.')


def report_with_links(title=None, summary=None, links=None, **kwargs):
    """
    Produces a generic report with a list of links

    Where:
        links : List[{href:str, title:str}]
    """
    if not isinstance(links, list):
        raise ValueError('Links needs to be a list.')
    if not all([(isinstance(link, dict) and 'title' in link and 'href' in link) for link in links]):
        raise ValueError('links items need to be dictionaries with href and title keys.')
    _validate_args(title, summary)
    report(title=title, detail={'summary': summary, 'links': links}, renderers=RenderWithLinks, **kwargs)


def report_with_remediation(title=None, summary=None, remediation=None, **kwargs):
    """
    Produces a generic report with a remediation

    Where:
        remediation : str
    """
    _validate_args(title, summary)
    report(title=title, detail={'summary': summary, 'remediation': remediation},
           renderers=RenderWithRemediation, **kwargs)


def report_generic(title=None, summary=None, **kwargs):
    """
    Produces a generic report with a summary
    """
    _validate_args(title, summary)
    report(title=title, detail={'summary': summary}, renderers=RenderGeneric, **kwargs)



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_generic
from leapp.models import Report, TargetRHSMInfo
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class ReportSetTargetRelease(Actor):
    """
    Reports that a release will be set in the subscription-manager after the upgrade.
    """

    name = 'report_set_target_release'
    consumes = (TargetRHSMInfo,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        info = next(self.consume(TargetRHSMInfo), None)
        if info and info.release:
            report_generic(
                title='The subscription-manager release is going to be set to {release}'.format(release=info.release),
                summary=(
                    'After the upgrade has completed the release of the subscription-manager will be set to {release}.'
                    ' This will ensure that you will receive and keep the version you choose to upgrade to.'
                ).format(release=info.release),
                severity='low'
            )



from leapp.models import Model, fields
from leapp.topics import SystemFactsTopic


class RepositoryData(Model):
    topic = SystemFactsTopic

    repoid = fields.String()
    name = fields.String()
    baseurl = fields.Nullable(fields.String())
    metalink = fields.Nullable(fields.String())
    mirrorlist = fields.Nullable(fields.String())
    enabled = fields.Boolean(default=True)
    additional_fields = fields.Nullable(fields.String())


class RepositoryFile(Model):
    topic = SystemFactsTopic

    file = fields.String()
    data = fields.List(fields.Model(RepositoryData))


class RepositoriesFacts(Model):
    topic = SystemFactsTopic

    repositories = fields.List(fields.Model(RepositoryFile))



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class RPM(Model):
    topic = SystemInfoTopic
    name = fields.String()
    epoch = fields.String()
    packager = fields.String()
    version = fields.String()
    release = fields.String()
    arch = fields.String()
    pgpsig = fields.String()
    repository = fields.Nullable(fields.String())


class InstalledRPM(Model):
    topic = SystemInfoTopic
    items = fields.List(fields.Model(RPM), default=[])


class InstalledRedHatSignedRPM(InstalledRPM):
    pass


class InstalledUnsignedRPM(InstalledRPM):
    pass



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class UdevAdmInfoData(Model):
    topic = SystemInfoTopic

    # Database export obtained by executing "udevadm info -e".
    db = fields.String()



import os
import re
import errno

from leapp.actors import Actor
from leapp.models import PersistentNetNamesFacts, PersistentNetNamesFactsInitramfs
from leapp.models import RenamedInterface, RenamedInterfaces, InitrdIncludes
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag

LINK_FILE_TEMPLATE = """# Generated by LEAPP
[Match]
MACAddress={}

[Link]
Name={}
"""


class PersistentNetNamesConfig(Actor):
    """
    Generate udev persistent network naming configuration

    This actor generates systemd-udevd link files for each physical ethernet interface present on RHEL-7
    in case we notice that interace name differs on RHEL-8. Link file configuration will assign RHEL-7 version of
    a name. Actors produces list of interfaces which changed name between RHEL-7 and RHEL-8.
    """

    name = 'persistentnetnamesconfig'
    consumes = (PersistentNetNamesFacts, PersistentNetNamesFactsInitramfs)
    produces = (RenamedInterfaces, InitrdIncludes)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)
    initrd_files = []

    def generate_link_file(self, interface):

        try:
            os.makedirs('/etc/systemd/network')
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        link_file = '/etc/systemd/network/10-leapp-{}.link'.format(interface.name)
        with open(link_file, 'w') as f:
            f.write(LINK_FILE_TEMPLATE.format(interface.mac, interface.name))

        self.initrd_files.append(link_file)

    def process(self):
        rhel7_ifaces = next(self.consume(PersistentNetNamesFacts)).interfaces
        rhel8_ifaces = next(self.consume(PersistentNetNamesFactsInitramfs)).interfaces

        sorted(rhel7_ifaces, key=lambda iface: iface.name)
        sorted(rhel8_ifaces, key=lambda iface: iface.name)

        rhel7_ifaces_map = {iface.mac: iface for iface in rhel7_ifaces}
        rhel8_ifaces_map = {iface.mac: iface for iface in rhel8_ifaces}

        renamed_interfaces = []

        if rhel7_ifaces != rhel8_ifaces:
            for iface in rhel7_ifaces:
                rhel7_name = rhel7_ifaces_map[iface.mac].name
                rhel8_name = rhel8_ifaces_map[iface.mac].name

                if rhel7_name != rhel8_name:
                    self.log.warning('Detected interface rename {} -> {}.'.format(rhel7_name, rhel8_name))

                    if re.search('eth[0-9]+', iface.name) is not None:
                        self.log.warning('Interface named using eth prefix, refusing to generate link file')
                        renamed_interfaces.append(RenamedInterface(**{'rhel7_name': rhel7_name,
                                                                      'rhel8_name': rhel8_name}))
                        continue
                    else:
                        # TODO(msekleta): we need to introduce another actor which will
                        # # install generated link files to RHEL-8 initrd image
                        self.generate_link_file(iface)

        self.produce(RenamedInterfaces(renamed=renamed_interfaces))
        self.produce(InitrdIncludes(files=self.initrd_files))



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_generic
from leapp.libraries.stdlib import run, CalledProcessError
from leapp.models import Authselect, AuthselectDecision
from leapp.reporting import Report
from leapp.tags import IPUWorkflowTag, ApplicationsPhaseTag, ExperimentalTag


class AuthselectApply(Actor):
    """
    Apply changes suggested by AuthselectScanner.

    If confirmed by admin in AuthselectDecision, call suggested authselect
    command to configure the system using this tool.
    """

    name = 'authselect_apply'
    consumes = (Authselect, AuthselectDecision,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ApplicationsPhaseTag, ExperimentalTag)

    def process(self):
        model = next(self.consume(Authselect))
        decision = next(self.consume(AuthselectDecision))

        if not decision.confirmed or model.profile is None:
            return

        command = ['authselect', 'select', '--force', model.profile] + model.features

        try:
            run(command)
        except CalledProcessError as err:
            report_generic(
                title='Authselect call failed.',
                summary=str(err)
            )
            return

        report_generic(
            title='System was converted to authselect.',
            summary='System was converted to authselect with the '
                    'following call: "{}"'.format(' '.join(command))
        )




def get_package_repository_data():
    """ Return dictionary mapping package name with repository from which it was installed """
    # import has to be inside the function to avoid troubles with non-existing
    # module in Python3 (where we do not need this function anymore)
    import yum
    yum_base = yum.YumBase()
    pkg_repos = {}
    for pkg in yum_base.doPackageLists().installed:
        pkg_repos[pkg.name] = pkg.ui_from_repo.lstrip('@')

    return pkg_repos



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_with_remediation
from leapp.libraries.common.rpms import has_package
from leapp.models import InstalledRedHatSignedRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckDosfstools(Actor):
    """
    Check if dosfstools is installed. If yes, write information about non-compatible changes.
    """

    name = 'checkdosfstools'
    consumes = (InstalledRedHatSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        if has_package(InstalledRedHatSignedRPM, 'dosfstools'):
            report_with_remediation(
                title='Dosfstools incompatible changes in the next major version',
                summary='The automatic alignment of data clusters that was added in 3.0.8 and broken for '
                        'FAT32 starting with 3.0.20 has been reinstated. If you need to create file systems '
                        'for finicky devices that have broken FAT implementations use the option -a to '
                        'disable alignment.\n'
                        'The fsck.fat now defaults to interactive repair mode which previously had to be '
                        'selected with the -r option.\n',
                remediation='Please update your scripts to be compatible with the changes.',
                severity='low')



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class XFSPresence(Model):
    topic = SystemInfoTopic

    present = fields.Boolean(default=False)
    without_ftype = fields.Boolean(default=False)



import re

from leapp.actors import Actor
from leapp.libraries.common import reporting
from leapp.models import KernelCmdlineArg, PersistentNetNamesFacts
from leapp.reporting import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class PersistentNetNamesDisable(Actor):
    """
    Disable systemd-udevd persistent network naming on machine with single eth0 NIC
    """

    name = 'persistentnetnamesdisable'
    consumes = (PersistentNetNamesFacts,)
    produces = (KernelCmdlineArg, Report)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def ethX_count(self, interfaces):
        ethX = re.compile('eth[0-9]+')
        count = 0

        for i in interfaces:
            if ethX.match(i.name):
                count = count + 1
        return count

    def single_eth0(self, interfaces):
        return len(interfaces) == 1 and interfaces[0].name == 'eth0'

    def disable_persistent_naming(self):
        self.log.info("Single eth0 network interface detected. Appending 'net.ifnames=0' to RHEL-8 kernel commandline")
        self.produce(KernelCmdlineArg(**{'key': 'net.ifnames', 'value': '0'}))

    def process(self):
        interfaces = next(self.consume(PersistentNetNamesFacts)).interfaces

        if self.single_eth0(interfaces):
            self.disable_persistent_naming()
        elif len(interfaces) > 1 and self.ethX_count(interfaces) > 0:
            reporting.report_generic(
                title='Unsupported network configuration',
                summary='Detected multiple network interfaces using unstable kernel names (e.g. eth0, eth1). '
                        'Upgrade process can not continue because stability of names can not be guaranteed. '
                        'Please read the article at https://access.redhat.com/solutions/4067471 for more information.',
                severity='high',
                flags=['inhibitor'])



import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import library
from leapp.models import BootContent


class run_mocked(object):
    def __init__(self):
        self.args = None

    def __call__(self, args, split=False):
        self.args = args


class write_to_file_mocked(object):
    def __init__(self):
        self.content = None

    def __call__(self, filename, content):
        self.content = content


def test_add_boot_entry(monkeypatch):
    def get_boot_file_paths_mocked():
        return '/abc', '/def'
    monkeypatch.setattr(library, 'get_boot_file_paths', get_boot_file_paths_mocked)
    monkeypatch.setenv('LEAPP_DEBUG', '1')
    monkeypatch.setattr(library, 'run', run_mocked())

    library.add_boot_entry()

    assert library.run.args == ['/usr/sbin/grubby',
                                '--add-kernel', '/abc',
                                '--initrd', '/def',
                                '--title', 'RHEL Upgrade Initramfs',
                                '--copy-default',
                                '--make-default',
                                '--args',
                                'debug enforcing=0 rd.plymouth=0 plymouth.enable=0']


def test_get_boot_file_paths(monkeypatch):
    # BootContent message available
    def consume_message_mocked(*models):
        yield BootContent(kernel_path='/ghi', initram_path='/jkl')
    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_message_mocked)

    kernel_path, initram_path = library.get_boot_file_paths()

    assert kernel_path == '/ghi' and initram_path == '/jkl'

    # No BootContent message available
    def consume_no_message_mocked(*models):
        yield None
    monkeypatch.setattr('leapp.libraries.stdlib.api.consume', consume_no_message_mocked)

    with pytest.raises(StopActorExecutionError):
        library.get_boot_file_paths()


def test_fix_grub_config_error(monkeypatch):
    monkeypatch.setattr(library, 'write_to_file', write_to_file_mocked())
    library.fix_grub_config_error('files/grub_test.wrong')

    with open('files/grub_test.fixed') as f:
        assert library.write_to_file.content == f.read()



import json
import os
from collections import namedtuple

from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp.libraries.common import reporting
from leapp.libraries.stdlib import api
from leapp.libraries.stdlib.config import is_verbose
from leapp.models import (InstalledRedHatSignedRPM, PESRpmTransactionTasks,
                          RepositoriesSetupTasks, RpmTransactionTasks, RepositoriesBlacklisted)

# FIXME: this mapping is not complete and will need to be manually updated frequently
REPOSITORIES_MAPPING = {
    'rhel8-appstream': 'rhel-8-for-x86_64-appstream-rpms',
    'rhel8-baseos': 'rhel-8-for-x86_64-baseos-rpms',
    'rhel8-crb': 'codeready-builder-for-rhel-8-x86_64-rpms'}

Event = namedtuple('Event', ['action',   # A string representing an event type (see EVENT_TYPES)
                             'in_pkgs',  # A dictionary with packages in format {<pkg_name>: <repository>}
                             'out_pkgs'  # A dictionary with packages in format {<pkg_name>: <repository>}
                             ])

EVENT_TYPES = ('Present', 'Removed', 'Deprecated', 'Replaced', 'Split', 'Merged', 'Moved', 'Renamed')


def pes_events_scanner(pes_json_filepath):
    """Entrypoint to the library"""
    installed_pkgs = get_installed_pkgs()
    transaction_configuration = get_transaction_configuration()
    events = get_events(pes_json_filepath)
    add_output_pkgs_to_transaction_conf(transaction_configuration, events)
    filtered_events = get_events_for_installed_pkgs_only(events, installed_pkgs)
    to_install, to_remove = process_events(filtered_events, installed_pkgs)
    filter_out_transaction_conf_pkgs(to_install, to_remove, transaction_configuration)
    produce_messages(to_install, to_remove)


def get_installed_pkgs():
    """
    Get installed Red Hat-signed packages.

    :return: Set of names of the installed Red Hat-signed packages
    """
    installed_pkgs = set()

    installed_rh_signed_rpm_msgs = api.consume(InstalledRedHatSignedRPM)
    installed_rh_signed_rpm_msg = next(installed_rh_signed_rpm_msgs, None)
    if list(installed_rh_signed_rpm_msgs):
        api.current_logger().warning('Unexpectedly received more than one InstalledRedHatSignedRPM message.')
    if not installed_rh_signed_rpm_msg:
        raise StopActorExecutionError('Cannot parse PES data properly due to missing list of installed packages',
                                      details={'Problem': 'Did not receive a message with installed Red Hat-signed '
                                                          'packages (InstalledRedHatSignedRPM)'})
    installed_pkgs.update([pkg.name for pkg in installed_rh_signed_rpm_msg.items])
    return installed_pkgs


def get_transaction_configuration():
    """
    Get pkgs to install, keep and remove from the user configuration files in /etc/leapp/transaction/.

    These configuration files have higher priority than PES data.
    :return: RpmTransactionTasks model instance
    """
    transaction_configuration = RpmTransactionTasks()

    for tasks in api.consume(RpmTransactionTasks):
        transaction_configuration.to_install.extend(tasks.to_install)
        transaction_configuration.to_remove.extend(tasks.to_remove)
        transaction_configuration.to_keep.extend(tasks.to_keep)
    return transaction_configuration


def get_events(pes_events_filepath):
    """
    Get all the events from the source JSON file exported from PES.

    :return: List of Event tuples, where each event contains event type and input/output pkgs
    """
    try:
        return parse_pes_events_file(pes_events_filepath)
    except (ValueError, KeyError):
        title = 'Missing/Invalid PES data file ({})'.format(pes_events_filepath)
        summary = 'Read documentation at: https://access.redhat.com/articles/3664871 for more information ' \
            'about how to retrieve the files'
        reporting.report_generic(title=title, summary=summary, severity='high', flags=['inhibitor'])
        raise StopActorExecution()


def parse_pes_events_file(path):
    """
    Parse JSON file returning PES events

    :return: List of Event tuples, where each event contains event type and input/output pkgs
    """
    if path is None or not os.path.isfile(path):
        raise ValueError('File {} not found'.format(path))
    with open(path) as f:
        data = json.load(f)

        if not isinstance(data, dict) or not data.get('packageinfo'):
            raise ValueError('Found PES data with invalid structure')

        return [parse_entry(entry) for entry in data['packageinfo']]


def parse_entry(entry):
    """
    Parse PES event data

    :param entry: A dict with the following structure:
                  {
                      'action': 3,
                      'id': 15,
                      'release': {
                          'z_stream': None,
                          'major_version': 8,
                          'tag': None,
                          'os_name': 'RHEL',
                          'minor_version': 0
                      },
                      'in_packageset': {
                          'set_id': 20,
                          'package': [
                              {
                                  'name': 'espeak',
                                  'repository': 'rhel7-optional'
                              }
                           ]
                      }
                      'out_packageset': {  # 'out_packageset' can be None
                          'set_id': 21,
                          'package': [
                              {
                                  'name': 'espeak-ng',
                                  'repository': 'rhel8-AppStream'
                              }
                          ]
                      },
                  }
    """
    action = parse_action(entry['action'])
    in_pkgs = parse_packageset(entry.get('in_packageset') or {})
    out_pkgs = parse_packageset(entry.get('out_packageset') or {})

    return Event(action, in_pkgs, out_pkgs)


def parse_action(action_id):
    """Get event type name based on PES event's action id"""
    if action_id < 0 or action_id >= len(EVENT_TYPES):
        raise ValueError('Found event with invalid action ID: {}'. format(action_id))

    return EVENT_TYPES[action_id]


def parse_packageset(packageset):
    """
    Get "input" or "output" packages and their repositories from each PES event.

    :return: A dictionary with packages in format {<pkg_name>: <repository>}
    """
    return {p['name']: p['repository'].lower() for p in packageset.get('package', [])}


def get_events_for_installed_pkgs_only(events, installed_rh_pkgs):
    """
    Get those PES events that have at least one of the event's "input" packages installed and signed by Red Hat.

    :param events: List of Event tuples, where each event contains event type and input/output pkgs
    :param installed_rh_pkgs: Set of names of the installed Red Hat-signed packages
    :return: List of Event tuples, not including those events whose input packages are not installed
    """

    filtered = []
    for event in events:
        if is_at_least_one_event_input_pkg_installed(installed_rh_pkgs, event.in_pkgs):
            filtered.append(event)

    return filtered


def is_at_least_one_event_input_pkg_installed(installed_rh_pkgs, event_in_pkgs):
    return installed_rh_pkgs.intersection(set(event_in_pkgs.keys()))


def process_events(events, installed_pkgs):
    """
    Process PES events to get lists of pkgs to install or to remove.

    Compare the "to install" list to the already installed pkgs to not mandate installation of what is already
    installed. Also, filter out installation of pkgs in blacklisted repos.

    :param events: List of Event tuples, not including those events with their "input" packages not installed
    :param installed_pkgs: Set of names of the installed Red Hat-signed packages
    :return: A tuple with two dicts {<pkg_name>: <repository>} - one with pkgs to install and one with pkgs to remove
    """
    # Dicts in format {<pkg_name>: <repository>}
    to_install = {}
    to_remove = {}

    for event in events:
        if event.action in ('Deprecated', 'Moved', 'Present'):
            # No need to mandate installation of these. If they are installed on the system, the default in the dnf
            # plugin is to add them to the "to upgrade" group, which is correct in this case.
            continue

        if event.action in ('Renamed', 'Replaced'):
            non_installed_out_pkgs = filter_out_installed_pkgs(event.out_pkgs, installed_pkgs)
            to_install.update(non_installed_out_pkgs)

        if event.action in ('Split', 'Merged'):
            non_installed_out_pkgs = filter_out_installed_pkgs(event.out_pkgs, installed_pkgs)
            to_install.update(non_installed_out_pkgs)
            in_pkgs_without_out_pkgs = filter_out_out_pkgs(event.in_pkgs, event.out_pkgs)
            to_remove.update(in_pkgs_without_out_pkgs)

        if event.action in ('Renamed', 'Replaced', 'Removed'):
            to_remove.update(event.in_pkgs)

    map_repositories(to_install)
    filter_out_pkgs_in_blacklisted_repos(to_install)

    return to_install, to_remove


def filter_out_installed_pkgs(event_out_pkgs, installed_pkgs):
    """Do not try to install those packages that are already installed."""
    return {k: v for k, v in event_out_pkgs.items() if k not in installed_pkgs}


def filter_out_out_pkgs(event_in_pkgs, event_out_pkgs):
    """
    In case of a Split or Merge PES events some of the "out" packages can be the same as the "in" packages.

    We don't want to remove those "in" packages that are among "out" packages. For example in case of a split of gdbm
    to gdbm and gdbm-libs, we would incorrectly mandate removing gdbm without this filter. But for example in case of
    a split of Cython to python2-Cython and python3-Cython, we will correctly mandate removing Cython.
    """
    return {k: v for k, v in event_in_pkgs.items() if k not in event_out_pkgs.keys()}


def filter_out_pkgs_in_blacklisted_repos(to_install):
    """Do not install packages that are available in blacklisted repositories"""
    # FIXME The to_install contains just a limited subset of packages - those that are *not* currently installed and
    # are to be installed. But we should also warn about the packages that *are* installed.
    blacklisted_pkgs = set()
    for pkg, repo in to_install.items():
        if repo in get_repositories_blacklisted():
            blacklisted_pkgs.add(pkg)
            del to_install[pkg]

    if blacklisted_pkgs:
        report_skipped_packages('packages will not be installed due to blacklisted repositories:',
                                blacklisted_pkgs)


def get_repositories_blacklisted():
    """Consume message and return a set of blacklisted repositories"""
    repos_blacklisted = set()
    for blacklist in api.consume(RepositoriesBlacklisted):
        repos_blacklisted.update(blacklist.repoids)
    return repos_blacklisted


def map_repositories(to_install):
    """Map repositories from PES data to RHSM repository id"""
    repo_not_mapped_pkgs = set()
    for pkg, repo in to_install.items():
        if repo not in REPOSITORIES_MAPPING:
            repo_not_mapped_pkgs.add(pkg)
            del to_install[pkg]
            continue

        to_install[pkg] = REPOSITORIES_MAPPING[repo]

    if repo_not_mapped_pkgs:
        report_skipped_packages('packages will not be installed due to not mapped repositories:',
                                repo_not_mapped_pkgs)


def report_skipped_packages(message, packages):
    """Generate report message about skipped packages"""
    title = 'Packages will not be installed'
    summary = '{} {}\n{}'.format(len(packages), message, '\n'.join(['- ' + p for p in packages]))
    reporting.report_generic(title=title, summary=summary, severity='high')
    if is_verbose():
        api.show_message(summary)


def add_output_pkgs_to_transaction_conf(transaction_configuration, events):
    """
    Add more packages for removal to transaction configuration if they can be derived as outputs of PES events.

    Output packages from an event are added to packages for removal only if all input packages are already there.

    :param events: List of Event tuples, where each event contains event type and input/output pkgs
    :param transaction_configuration: RpmTransactionTasks model instance with pkgs to install, keep and remove based
                                      on the user configuration files
    """
    message = 'Marking packages for removal as a result of events:\n'

    for event in events:
        if event.action in ('Split', 'Merged', 'Replaced', 'Renamed'):
            if all([pkg in transaction_configuration.to_remove for pkg in event.in_pkgs]):
                transaction_configuration.to_remove.extend(event.out_pkgs)
                message += '- [{action}] {ins} -> {outs}\n'.format(
                    action=event.action,
                    ins=', '.join(event.in_pkgs.keys()),
                    outs=', '.join(event.out_pkgs.keys())
                )

    api.current_logger().debug(message)


def filter_out_transaction_conf_pkgs(to_install, to_remove, transaction_configuration):
    """
    Filter out those PES events that conflict with the higher priority transaction configuration files.

    :param to_install: A dict {<pkg_name>: <repository>} with pkgs to install based on PES data
    :param to_remove: A dict {<pkg_name>: <repository>} with pkgs to remove based on PES data
    :param transaction_configuration: RpmTransactionTasks model instance with pkgs to install, keep and remove based
                                      on the user configuration files
    """
    # The two lists below are for debug message purposes only
    pes_based_pkgs_not_to_be_installed = []
    pes_based_pkgs_not_to_be_removed = []

    for pes_based_pkg_to_install in to_install.keys():
        if pes_based_pkg_to_install in transaction_configuration.to_remove:
            to_install.pop(pes_based_pkg_to_install)
            pes_based_pkgs_not_to_be_installed.append(pes_based_pkg_to_install)
    for pes_based_pkg_to_remove in to_remove.keys():
        if pes_based_pkg_to_remove in transaction_configuration.to_install + transaction_configuration.to_keep:
            to_remove.pop(pes_based_pkg_to_remove)
            pes_based_pkgs_not_to_be_removed.append(pes_based_pkg_to_remove)

    if pes_based_pkgs_not_to_be_installed:
        api.current_logger().debug('The following packages will not be installed because of the'
                                   ' /etc/leapp/transaction/to_remove transaction configuration file:'
                                   '\n- ' + '\n- '.join(pes_based_pkgs_not_to_be_installed))
    if pes_based_pkgs_not_to_be_removed:
        api.current_logger().debug('The following packages will not be removed because of the to_keep and to_install'
                                   ' transaction configuration files in /etc/leapp/transaction/:'
                                   '\n- ' + '\n- '.join(pes_based_pkgs_not_to_be_removed))


def produce_messages(to_install, to_remove):
    to_install_pkgs = set(to_install.keys())
    to_remove_pkgs = set(to_remove.keys())
    to_enable_repos = set(to_install.values())

    if to_install_pkgs or to_remove_pkgs:
        api.produce(PESRpmTransactionTasks(to_install=list(to_install_pkgs),
                                           to_remove=list(to_remove_pkgs)))

    if to_enable_repos:
        api.produce(RepositoriesSetupTasks(to_enable=list(to_enable_repos)))



import logging
import sys

from leapp.repository.scan import find_and_scan_repositories


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, filename='/dev/null')
    logger = logging.getLogger('run_pytest.py')

    BASE_REPO = 'repos'
    repos = find_and_scan_repositories(BASE_REPO, include_locals=True)
    repos.load()
    if len(sys.argv) > 1:
        actor = repos.lookup_actor(sys.argv[1])
        if actor:
            print(actor.full_path)
        else:
            sys.stdout.write('No actor found for search "{}"\n'.format(sys.argv[1]))
            sys.exit(1)
    else:
        sys.stdout.write('Missing commandline argument\n')
        sys.exit(1)



from leapp.libraries.common import reporting
from leapp.models import OSReleaseFacts


def get_os_release_info(path):
    ''' Retrieve data about System OS release from provided file '''
    data = {}
    try:
        with open(path) as f:
            data = dict(l.strip().split('=', 1) for l in f.readlines() if '=' in l)
    except IOError as e:
        reporting.report_generic(
            title='Error while collecting system OS facts',
            summary=str(e),
            severity='high',
            flags=['inhibitor'])
        return None

    return OSReleaseFacts(
        release_id=data.get('ID', '').strip('"'),
        name=data.get('NAME', '').strip('"'),
        pretty_name=data.get('PRETTY_NAME', '').strip('"'),
        version=data.get('VERSION', '').strip('"'),
        version_id=data.get('VERSION_ID', '').strip('"'),
        variant=data.get('VARIANT', '').strip('"') or None,
        variant_id=data.get('VARIANT_ID', '').strip('"') or None
    )



from leapp.models import PCIDevices, PCIDevice
from leapp.libraries.actor.pcidevicesscanner import parse_pci_devices, produce_pci_devices


def test_parse_pci_devices(current_actor_libraries):
    devices = [
        '00:00.0 "Host bridge" "Intel Corporation" "440FX - 82441FX PMC [Natoma]" -r02 "Red Hat, Inc." '
        '"Qemu virtual machine"',
        '00:01.0 "ISA bridge" "Intel Corporation" "82371SB PIIX3 ISA [Natoma/Triton II]" "Red Hat, Inc." '
        '"Qemu virtual machine"',
        '00:01.1 "IDE interface" "Intel Corporation" "82371SB PIIX3 IDE [Natoma/Triton II]" -p80 "Red Hat, '
        'Inc." "Qemu virtual machine"']

    output = parse_pci_devices(devices)
    assert isinstance(output, list)
    assert len(output) == 3

    dev = output.pop()
    assert dev.slot == '00:01.1'
    assert dev.dev_cls == 'IDE interface'
    assert dev.vendor == 'Intel Corporation'
    assert dev.name == '82371SB PIIX3 IDE [Natoma/Triton II]'
    assert dev.subsystem_vendor == 'Red Hat, Inc.'
    assert dev.subsystem_name == 'Qemu virtual machine'
    assert dev.progif == '80'

    dev = output.pop()
    assert dev.slot == '00:01.0'
    assert dev.dev_cls == 'ISA bridge'
    assert dev.vendor == 'Intel Corporation'
    assert dev.name == '82371SB PIIX3 ISA [Natoma/Triton II]'
    assert dev.subsystem_vendor == 'Red Hat, Inc.'
    assert dev.subsystem_name == 'Qemu virtual machine'

    dev = output.pop()
    assert dev.slot == '00:00.0'
    assert dev.dev_cls == 'Host bridge'
    assert dev.vendor == 'Intel Corporation'
    assert dev.name == '440FX - 82441FX PMC [Natoma]'
    assert dev.subsystem_vendor == 'Red Hat, Inc.'
    assert dev.subsystem_name == 'Qemu virtual machine'
    assert dev.rev == '02'


def test_parse_empty_list(current_actor_libraries):
    output = parse_pci_devices([])
    assert isinstance(output, list)
    assert not output


def test_parse_unknown_optional_parameter(current_actor_libraries):
    devices = ['00:01.1 -a01 "IDE interface" -b02 "Intel Corporation" -c03 "82371SB PIIX3 IDE [Natoma/Triton II]" '
               '-p80 "Red Hat, Inc." -d04 "Qemu virtual machine"']

    output = parse_pci_devices(devices)
    assert isinstance(output, list)
    assert len(output) == 1

    dev = output.pop()
    assert dev.slot == '00:01.1'
    assert dev.dev_cls == 'IDE interface'
    assert dev.vendor == 'Intel Corporation'
    assert dev.name == '82371SB PIIX3 IDE [Natoma/Triton II]'
    assert dev.subsystem_vendor == 'Red Hat, Inc.'
    assert dev.subsystem_name == 'Qemu virtual machine'
    assert dev.progif == '80'


def test_produce_pci_devices(current_actor_libraries):
    output = []

    def fake_producer(*args):
        output.extend(args)

    devices = [
        PCIDevice(
            slot='00:00.0',
            dev_cls='Host bridge',
            vendor='Intel Corporation',
            name='440FX - 82441FX PMC [Natoma]',
            subsystem_vendor='Red Hat, Inc.',
            subsystem_name='Qemu virtual machine',
            rev='02'),
        PCIDevice(
            slot='00:01.0',
            dev_cls='ISA bridge',
            vendor='Intel Corporation',
            name='82371SB PIIX3 ISA [Natoma/Triton II]',
            subsystem_vendor='Red Hat, Inc.',
            subsystem_name='Qemu virtual machine'),
        PCIDevice(
            slot='00:01.1',
            dev_cls='IDE interface',
            vendor='Intel Corporation',
            name='82371SB PIIX3 IDE [Natoma/Triton II]',
            subsystem_vendor='Red Hat, Inc.',
            subsystem_name='Qemu virtual machine',
            progif='80'),
    ]

    produce_pci_devices(fake_producer, devices)
    assert len(output) == 1
    assert len(output[0].devices) == 3


def test_produce_no_devices(current_actor_libraries):
    output = []

    def fake_producer(*args):
        output.extend(args)

    produce_pci_devices(fake_producer, [])
    assert len(output) == 1
    assert not output[0].devices


def test_actor_execution(current_actor_context):
    current_actor_context.run()
    assert current_actor_context.consume(PCIDevices)



import os
import shutil

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import dnfplugin, mounting
from leapp.libraries.stdlib import api
from leapp.models import (BootContent, RequiredUpgradeInitramPackages, TargetUserSpaceInfo, UpgradeDracutModule,
                          UsedTargetRepositories)

INITRAM_GEN_SCRIPT_NAME = 'generate-initram.sh'


def _reinstall_leapp_repository_hint():
    """
    Convenience function for creating a detail for StopActorExecutionError with a hint to reinstall the
    leapp-repository package
    """
    return {
        'hint': 'Try to reinstall the `leapp-repository` package.'
    }


def copy_dracut_modules(context, modules):
    """
    Copies our dracut modules into the target userspace.
    """
    try:
        shutil.rmtree(context.full_path('/dracut'))
    except EnvironmentError:
        pass
    for module in modules:
        try:
            context.copytree_to(module.module_path, os.path.join('/dracut', os.path.basename(module.module_path)))
        except shutil.Error as e:
            api.current_logger().error('Failed to copy dracut module "{name}" from "{source}" to "{target}"'.format(
                name=module.name, source=module.module_path, target=context.full_path('/dracut')), exc_info=True)
            raise StopActorExecutionError(
                message='Failed to install dracut modules required in the initram. Error: {}'.format(str(e))
            )


def _get_dracut_modules():
    return list(api.consume(UpgradeDracutModule))


def install_initram_deps(context):
    used_repos = api.consume(UsedTargetRepositories)
    target_userspace_info = next(api.consume(TargetUserSpaceInfo), None)

    packages = set()
    for message in api.consume(RequiredUpgradeInitramPackages):
        packages.update(message.packages)
    dnfplugin.install_initramdisk_requirements(
        packages=packages, target_userspace_info=target_userspace_info, used_repos=used_repos)


def generate_initram_disk(context):
    """
    Function to actually execute the initram creation.
    """
    generator_script = api.get_actor_file_path(INITRAM_GEN_SCRIPT_NAME)
    if not generator_script:
        raise StopActorExecutionError(
            message='Mandatory script to generate initram not available.',
            details=_reinstall_leapp_repository_hint()
        )
    modules = _get_dracut_modules()
    copy_dracut_modules(context, modules)
    context.copy_to(generator_script, os.path.join('/', INITRAM_GEN_SCRIPT_NAME))
    install_initram_deps(context)
    context.call([
        '/bin/sh', '-c',
        'LEAPP_ADD_DRACUT_MODULES={modules} {cmd}'.format(
            modules=','.join([mod.name for mod in modules]),
            cmd=os.path.join('/', INITRAM_GEN_SCRIPT_NAME))
    ])
    copy_boot_files(context)


def copy_boot_files(context):
    """
    Function to copy the generated initram and corresponding kernel to /boot - Additionally produces a BootContent
    message with their location.
    """
    kernel = 'vmlinuz-upgrade.x86_64'
    initram = 'initramfs-upgrade.x86_64.img'
    content = BootContent(
        kernel_path=os.path.join('/boot', kernel),
        initram_path=os.path.join('/boot', initram)
    )

    context.copy_from(os.path.join('/artifacts', kernel), content.kernel_path)
    context.copy_from(os.path.join('/artifacts', initram), content.initram_path)

    api.produce(content)


def process():
    userspace_info = next(api.consume(TargetUserSpaceInfo), None)

    with mounting.NspawnActions(base_dir=userspace_info.path) as context:
        generate_initram_disk(context)



import pytest

from leapp.exceptions import StopActorExecution
from leapp.libraries.actor.library import scan_repositories
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import produce_mocked, report_generic_mocked
from leapp.libraries.stdlib import api
from leapp.models import RepositoriesMap, RepositoryMap


def test_scan_valid_file(monkeypatch):
    monkeypatch.setattr(api, 'produce', produce_mocked())
    scan_repositories('files/tests/sample01.csv')

    assert api.produce.called == 1
    assert len(api.produce.model_instances) == 1
    assert isinstance(api.produce.model_instances[0], RepositoriesMap)
    assert len(api.produce.model_instances[0].repositories) == 4
    assert isinstance(api.produce.model_instances[0].repositories[0], RepositoryMap)
    assert api.produce.model_instances[0].repositories[0].from_id == 'FromRepo01'
    assert api.produce.model_instances[0].repositories[0].repo_type == 'rpm'
    assert isinstance(api.produce.model_instances[0].repositories[3], RepositoryMap)
    assert api.produce.model_instances[0].repositories[3].to_id == 'ToRepo04'
    assert api.produce.model_instances[0].repositories[3].arch == 'ppc64le'


def test_repository_mapping_file_not_found(monkeypatch):
    def file_not_exists(_filepath):
        return False
    monkeypatch.setattr('os.path.isfile', file_not_exists)
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    with pytest.raises(StopActorExecution):
        scan_repositories('/etc/leapp/files/repomap.csv')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']
    assert 'Repositories map file not found' in reporting.report_generic.report_fields['title']


def test_scan_empty_file(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    with pytest.raises(StopActorExecution):
        scan_repositories('files/tests/sample02.csv')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']
    assert 'Repositories map file is invalid' in reporting.report_generic.report_fields['title']


def test_scan_invalid_file_txt(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    with pytest.raises(StopActorExecution):
        scan_repositories('files/tests/sample03.csv')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']
    assert 'Repositories map file is invalid' in reporting.report_generic.report_fields['title']


def test_scan_invalid_file_csv(monkeypatch):
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())
    with pytest.raises(StopActorExecution):
        scan_repositories('files/tests/sample04.csv')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']
    assert 'Repositories map file is invalid' in reporting.report_generic.report_fields['title']



import os

import pytest

from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp.libraries.actor import library
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import produce_mocked, report_generic_mocked
from leapp.libraries.stdlib import api
from leapp.models import OSReleaseFacts

SUPPORTED_VERSION = {'rhel': ['7.5', '7.6']}


def test_skip_check(monkeypatch):
    monkeypatch.setattr(os, "getenv", lambda _unused: True)
    monkeypatch.setattr(reporting, "report_generic", report_generic_mocked())

    assert library.skip_check()
    assert reporting.report_generic.called == 1
    assert 'Skipped OS release check' in reporting.report_generic.report_fields['title']
    assert reporting.report_generic.report_fields['severity'] == 'high'
    assert 'flags' not in reporting.report_generic.report_fields


def test_no_skip_check(monkeypatch):
    monkeypatch.setattr(os, "getenv", lambda _unused: False)
    monkeypatch.setattr(reporting, "report_generic", report_generic_mocked())

    assert not library.skip_check()
    assert reporting.report_generic.called == 0


def test_no_facts(monkeypatch):
    def os_release_mocked(*models):
        yield None

    monkeypatch.setattr(api, "consume", os_release_mocked)
    with pytest.raises(StopActorExecutionError):
        library.check_os_version(SUPPORTED_VERSION)


def create_os_release(release_id, version_id=7.6):
    return OSReleaseFacts(
        release_id=release_id,
        name='test',
        pretty_name='test {}'.format(version_id),
        version='Some Test {}'.format(version_id),
        version_id=version_id,
        variant=None,
        variant_id=None,
    )


def test_not_supported_id(monkeypatch):
    def os_release_mocked(*models):
        yield create_os_release('rhel', '7.7')

    monkeypatch.setattr(api, "consume", os_release_mocked)
    monkeypatch.setattr(reporting, "report_generic", report_generic_mocked())

    library.check_os_version(SUPPORTED_VERSION)
    assert reporting.report_generic.called == 1
    assert 'Unsupported OS version' in reporting.report_generic.report_fields['title']
    assert 'flags' in reporting.report_generic.report_fields
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']


def test_not_supported_release(monkeypatch):
    def os_release_mocked(*models):
        yield create_os_release('unsupported', SUPPORTED_VERSION['rhel'][0])

    monkeypatch.setattr(api, "consume", os_release_mocked)
    monkeypatch.setattr(reporting, "report_generic", report_generic_mocked())

    library.check_os_version(SUPPORTED_VERSION)
    assert reporting.report_generic.called == 1
    assert 'Unsupported OS' in reporting.report_generic.report_fields['title']
    assert 'flags' in reporting.report_generic.report_fields
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']


def test_supported_release(monkeypatch):
    def os_mocked_first_release(*models):
        yield create_os_release('rhel', SUPPORTED_VERSION['rhel'][0])

    def os_mocked_second_release(*models):
        yield create_os_release('rhel', SUPPORTED_VERSION['rhel'][1])

    monkeypatch.setattr(api, "consume", os_mocked_first_release)
    monkeypatch.setattr(reporting, "report_generic", report_generic_mocked())

    library.check_os_version(SUPPORTED_VERSION)
    monkeypatch.setattr(api, "consume", os_mocked_second_release)
    library.check_os_version(SUPPORTED_VERSION)
    assert reporting.report_generic.called == 0


def test_invalid_versions(monkeypatch):
    def os_release_mocked(*models):
        yield create_os_release('rhel', '7.6')

    monkeypatch.setattr(api, "consume", os_release_mocked)
    monkeypatch.setattr(reporting, "report_generic", report_generic_mocked())

    with pytest.raises(StopActorExecution):
        library.check_os_version('string')
    with pytest.raises(StopActorExecution):
        library.check_os_version(None)

    library.check_os_version({})
    assert reporting.report_generic.called == 1
    with pytest.raises(StopActorExecutionError):
        library.check_os_version({'rhel': None})



from leapp.models import Model, fields
from leapp.topics import TransactionTopic


class RepositoryMap(Model):
    """
    Mapping between repositories to be used during upgrade.

    Mapping of current system repository to target system version repositories to determine which
    ones should be enabled for the correct upgrade process.
    """

    topic = TransactionTopic

    from_id = fields.String()
    to_id = fields.String()
    from_minor_version = fields.String()
    to_minor_version = fields.String()
    arch = fields.String()
    repo_type = fields.StringEnum(choices=['rpm', 'srpm', 'debuginfo'])


class RepositoriesMap(Model):
    topic = TransactionTopic

    repositories = fields.List(fields.Model(RepositoryMap), default=[])



from leapp.actors import Actor
from leapp.models import StorageInfo
from leapp.reporting import Report
from leapp.libraries.common.reporting import report_with_remediation
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class InhibitWhenLuks(Actor):
    """
    Check if any encrypted partitions is in use. If yes, inhibit the upgrade process.

    Upgrading system with encrypted partition is not supported.
    """

    name = 'check_luks_and_inhibit'
    consumes = (StorageInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for storage_info in self.consume(StorageInfo):
            for blk in storage_info.lsblk:
                if blk.tp == 'crypt':
                    report_with_remediation(
                        title='LUKS encrypted partition detected',
                        summary='Upgrading system with encrypted partitions is not supported',
                        remediation='If the encrypted partition is not system one and the system '
                                    'is not depending on it, you can remove/blacklist it from '
                                    'the system',
                        severity='high',
                        flags=['inhibitor'])
                    break



import logging

from leapp.libraries.actor.scanner import load_tasks_file, load_tasks


def test_load_tasks(tmpdir):
    tmpdir.join('to_install').write('a\n b\n  c \n\n\nc\na\nc\nb')
    tmpdir.join('to_keep').write('a\n b\n  c \n\n\nc\na\nc\nb')
    tmpdir.join('to_remove').write('a\n b\n  c \n\n\nc\na\nc\nb')
    m = load_tasks(tmpdir.strpath, logging)
    assert set(m.to_install) == set(['a', 'b', 'c'])
    assert set(m.to_keep) == set(['a', 'b', 'c'])
    assert set(m.to_remove) == set(['a', 'b', 'c'])


def test_load_tasks_file(tmpdir):
    f = tmpdir.join('to_install')
    f.write('a\n b\n  c \n\n\nc\na\nc\nb')
    assert set(load_tasks_file(f.strpath, logging)) == set(['a', 'b', 'c'])
    f = tmpdir.join('to_keep')
    f.write(' ')
    assert set(load_tasks_file(f.strpath, logging)) == set([])



from leapp.actors import Actor
from leapp.libraries.actor import library
from leapp.libraries.common.reporting import report_generic
from leapp.models import BrlttyMigrationDecision
from leapp.reporting import Report
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class MigrateBrltty(Actor):
    """
    Migrate brltty configuration files.
    """

    name = 'migrate_brltty'
    consumes = (BrlttyMigrationDecision,)
    produces = (Report,)
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        for decision in self.consume(BrlttyMigrationDecision):
            report_summary = ''
            library.migrate_file(decision.migrate_file, decision.migrate_bt, decision.migrate_espeak)
            if decision.migrate_bt:
                report_summary = 'Unsupported aliases for bluetooth devices (\'bth:\' and \'bluez:\') was '
                report_summary += 'renamed to \'bluetooth:\' in {}'
                report_summary = report_summary.format(', '.join(decision.migrate_file))
            if decision.migrate_espeak:
                if report_summary:
                    report_summary += '\n'
                report_summary += 'eSpeak speech driver was switched to eSpeak-NG in {}'
                report_summary = report_summary.format(', '.join(decision.migrate_file))
            if decision.migrate_bt or decision.migrate_espeak:
                report_generic(
                    title='brltty configuration files migrated',
                    summary=report_summary,
                    severity='low')



from leapp.libraries.actor import library
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import UsedTargetRepositories, UsedTargetRepository


class run_mocked(object):
    def __init__(self, raise_err=False):
        self.called = 0
        self.args = []
        self.raise_err = raise_err

    def __call__(self, *args):
        self.called += 1
        self.args.append(args)
        if self.raise_err:
            raise CalledProcessError(
                message='A Leapp Command Error occured.',
                command=args,
                result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'}
            )


class logger_mocked(object):
    def __init__(self):
        self.errmsg = None

    def error(self, *args):
        self.errmsg = args

    def __call__(self):
        return self


def construct_UTRepo_consume(repoids):
    repos = [UsedTargetRepository(repoid=repoid) for repoid in repoids]
    return lambda *x: (x for x in (UsedTargetRepositories(repos=repos),))


def test_is_crb_used_false(monkeypatch):
    inputs = ([], ['some-name'], ['some-name', 'another-repo'])
    for i in inputs:
        monkeypatch.setattr(api, 'consume', construct_UTRepo_consume(i))
        assert not library._is_crb_used()


def test_is_crb_used_true(monkeypatch):
    inputs = ([library.CRB_REPOID], ['some-name', library.CRB_REPOID])
    for i in inputs:
        monkeypatch.setattr(api, 'consume', construct_UTRepo_consume(i))
        assert library._is_crb_used()


def test_process_setrepo(monkeypatch):
    monkeypatch.setattr(library, 'run', run_mocked())
    monkeypatch.setattr(library, '_is_crb_used', lambda: True)
    library.process()
    assert library.run.called
    assert 'subscription-manager' in library.run.args[0][0]
    assert library.CRB_REPOID in library.run.args[0][0]


def test_process_donothing(monkeypatch):
    monkeypatch.setattr(library, 'run', run_mocked())
    monkeypatch.setattr(library, '_is_crb_used', lambda: False)
    library.process()
    assert not library.run.called


def test_process_fail(monkeypatch):
    monkeypatch.setattr(library, 'run', run_mocked(raise_err=True))
    monkeypatch.setattr(api, 'current_logger', logger_mocked())
    monkeypatch.setattr(library, '_is_crb_used', lambda: True)
    library.process()
    assert library.run.called
    assert api.current_logger.errmsg



from leapp.actors import Actor
from leapp.libraries.actor.library import add_boot_entry, fix_grub_config_error
from leapp.models import BootContent, GrubConfigError
from leapp.tags import InterimPreparationPhaseTag, IPUWorkflowTag


class AddUpgradeBootEntry(Actor):
    """
    Add new boot entry for Leapp provided initramfs.

    Using new boot entry, Leapp can continue the upgrade process in the initramfs after reboot
    """

    name = 'add_upgrade_boot_entry'
    consumes = (BootContent, GrubConfigError)
    produces = ()
    tags = (IPUWorkflowTag, InterimPreparationPhaseTag)

    def process(self):
        grub_config_error_detected = next(self.consume(GrubConfigError), GrubConfigError()).error_detected
        if grub_config_error_detected:
            fix_grub_config_error('/etc/default/grub')

        add_boot_entry()



from leapp.libraries.common import rhsm
from leapp.models import Report, SourceRHSMInfo


def test_sku_report_skipped(monkeypatch, current_actor_context):
    with monkeypatch.context() as context:
        context.setenv('LEAPP_DEVEL_SKIP_RHSM', '1')
        current_actor_context.feed(SourceRHSMInfo(attached_skus=[]))
        current_actor_context.run()
        assert not list(current_actor_context.consume(Report))


def test_sku_report_has_skus(monkeypatch, current_actor_context):
    with monkeypatch.context() as context:
        context.setenv('LEAPP_DEVEL_SKIP_RHSM', '0')
        current_actor_context.feed(SourceRHSMInfo(attached_skus=['testing-sku']))
        current_actor_context.run()
        assert not list(current_actor_context.consume(Report))


def test_sku_report_has_no_skus(monkeypatch, current_actor_context):
    with monkeypatch.context() as context:
        context.setenv('LEAPP_DEVEL_SKIP_RHSM', '0')
        current_actor_context.feed(SourceRHSMInfo(attached_skus=[]))
        current_actor_context.run()
        reports = list(current_actor_context.consume(Report))
        assert reports and len(reports) == 1
        assert reports[0].severity == 'high'
        assert reports[0].title == 'The system is not registered or subscribed.'



import os

import leapp.libraries.stdlib


class run_mocked(object):
    def __init__(self):
        self.called = 0
        self.args = None

    def __call__(self, args, split=False):
        self.called += 1
        self.args = args
        return {'stdout': ''}


def _get_tool_path(name):
    for directory in os.getenv('LEAPP_COMMON_TOOLS', '').split(':'):
        full_path = os.path.join(directory, name)
        if os.path.isfile(full_path):
            return full_path
    return None


def test_prepare_yum_config(monkeypatch):
    run = run_mocked()
    with monkeypatch.context() as context:
        context.setattr(leapp.libraries.stdlib, 'run', run)
        context.setattr(leapp.libraries.stdlib.api, 'get_tool_path', _get_tool_path)
        from leapp.libraries.common import utils  # Needed locally to allow monkey patching to actually work
        utils.apply_yum_workaround()
    assert run.called == 1
    assert os.path.basename(run.args[-1]) == 'handleyumconfig'



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class EnvVar(Model):
    topic = SystemInfoTopic

    name = fields.String()
    value = fields.String()


class OSRelease(Model):
    topic = SystemInfoTopic

    release_id = fields.String()
    name = fields.String()
    pretty_name = fields.String()
    version = fields.String()
    version_id = fields.String()
    variant = fields.Nullable(fields.String())  # isn't specified on some systems
    variant_id = fields.Nullable(fields.String())  # same as above


class Version(Model):
    topic = SystemInfoTopic

    source = fields.String()
    target = fields.String()


class IPUConfig(Model):
    """
    IPU workflow configuration model
    """
    topic = SystemInfoTopic

    leapp_env_vars = fields.List(fields.Model(EnvVar), default=[])
    os_release = fields.Model(OSRelease)
    version = fields.Model(Version)
    architecture = fields.String()



import contextlib
import itertools
import json
import os
import shutil

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import guards, mounting, overlaygen, utils
from leapp.libraries.stdlib import CalledProcessError, api, config

DNF_PLUGIN_NAME = 'rhel_upgrade.py'
DNF_PLUGIN_PATH = os.path.join('/lib/python3.6/site-packages/dnf-plugins', DNF_PLUGIN_NAME)
DNF_PLUGIN_DATA_NAME = 'dnf-plugin-data.txt'
DNF_PLUGIN_DATA_PATH = os.path.join('/var/lib/leapp', DNF_PLUGIN_DATA_NAME)
DNF_PLUGIN_DATA_LOG_PATH = os.path.join('/var/log/leapp', DNF_PLUGIN_DATA_NAME)
DNF_DEBUG_DATA_PATH = '/var/log/leapp/dnf-debugdata/'


def install(target_basedir):
    """
    Installs our plugin to the DNF plugins.
    """
    try:
        shutil.copy2(
            api.get_file_path(DNF_PLUGIN_NAME),
            os.path.join(target_basedir, DNF_PLUGIN_PATH.lstrip('/')))
    except EnvironmentError as e:
        api.current_logger().debug('Failed to install DNF plugin', exc_info=True)
        raise StopActorExecutionError(
            message='Failed to install DNF plugin. Error: {}'.format(str(e))
        )


def build_plugin_data(target_repoids, debug, test, tasks):
    """
    Generates a dictionary with the DNF plugin data.
    """
    # get list of repo IDs of target repositories that should be used for upgrade
    data = {
        'pkgs_info': {
            'local_rpms': [os.path.join('/installroot', pkg.lstrip('/')) for pkg in tasks.local_rpms],
            'to_install': [pkg for pkg in tasks.to_install],
            'to_remove': [pkg for pkg in tasks.to_remove],
            #            'to_upgrade': [pkg for pkg in tasks.to_upgrade]
        },
        'dnf_conf': {
            'allow_erasing': True,
            'best': True,
            'debugsolver': debug,
            'disable_repos': True,
            'enable_repos': target_repoids,
            'gpgcheck': False,
            'platform_id': 'platform:el8',
            'releasever': '8',
            'installroot': '/installroot',
            'test_flag': test
        }
    }
    return data


def create_config(context, target_repoids, debug, test, tasks):
    """
    Creates the configuration data file for our DNF plugin.
    """
    context.makedirs(os.path.dirname(DNF_PLUGIN_DATA_PATH), exists_ok=True)
    with context.open(DNF_PLUGIN_DATA_PATH, 'w+') as f:
        config_data = build_plugin_data(target_repoids=target_repoids, debug=debug, test=test, tasks=tasks)
        json.dump(config_data, f, sort_keys=True, indent=2)


def backup_config(context):
    """
    Backs up the configuration data used for the plugin.
    """
    context.copy_from(DNF_PLUGIN_DATA_PATH, DNF_PLUGIN_DATA_LOG_PATH)


def backup_debug_data(context):
    """
    Performs the backup of DNF debug data
    """
    if config.is_debug():
        # The debugdata is a folder generated by dnf when using the --debugsolver dnf option. We switch on the
        # debug_solver dnf config parameter in our rhel-upgrade dnf plugin when LEAPP_DEBUG env var set to 1.
        try:
            context.copytree_from('/debugdata', DNF_DEBUG_DATA_PATH)
        except OSError as e:
            api.current_logger().warn('Failed to copy debugdata. Message: {}'.format(str(e)), exc_info=True)


def _transaction(context, stage, target_repoids, tasks, test=False):
    """
    Perform the actual DNF rpm download via our DNF plugin
    """

    create_config(context=context, target_repoids=target_repoids, debug=config.is_debug(), test=test, tasks=tasks)
    backup_config(context=context)

    with guards.guarded_execution(guards.connection_guard(), guards.space_guard()):
        try:
            context.call(
                cmd=[
                    '/usr/bin/dnf',
                    'rhel-upgrade',
                    stage,
                    DNF_PLUGIN_DATA_PATH
                ],
                callback_raw=utils.logging_handler
            )
        except OSError as e:
            api.current_logger().error('Could not call dnf command: Message: %s', str(e), exc_info=True)
            raise StopActorExecutionError(
                message='Failed to execute dnf. Reason: {}'.format(str(e))
            )
        except CalledProcessError as e:
            api.current_logger().error('DNF execution failed: ')
            raise StopActorExecutionError(
                message='DNF execution failed with non zero exit code.\nSTDOUT:\n{stdout}\nSTDERR:\n{stderr}'.format(
                    stdout=e.stdout, stderr=e.stderr)
            )
        finally:
            if stage == 'check':
                backup_debug_data(context=context)


@contextlib.contextmanager
def _prepare_transaction(used_repos, target_userspace_info, binds=()):
    """ Creates the transaction environment needed for the target userspace DNF execution  """
    target_repoids = set()
    for message in used_repos:
        target_repoids.update([repo.repoid for repo in message.repos])
    with mounting.NspawnActions(base_dir=target_userspace_info.path, binds=binds) as context:
        yield context, list(target_repoids), target_userspace_info


def install_initramdisk_requirements(packages, target_userspace_info, used_repos):
    """
    Performs the installation of packages into the initram disk
    """
    with _prepare_transaction(used_repos=used_repos,
                              target_userspace_info=target_userspace_info) as (context, target_repoids, _unused):
        repos_opt = [['--enablerepo', repo] for repo in target_repoids]
        repos_opt = list(itertools.chain(*repos_opt))
        context.call(
            [
                'dnf',
                'install',
                '-y',
                '--nogpgcheck',
                '--setopt=module_platform_id=platform:el8',
                '--setopt=keepcache=1',
                '--disablerepo', '*'
            ] + repos_opt + [package for package in packages]
        )


def perform_transaction_install(target_userspace_info, used_repos, tasks):
    """
    Performs the actual installation with the DNF rhel-upgrade plugin using the target userspace
    """

    # These bind mounts are performed by systemd-nspawn --bind parameters
    bind_mounts = [
        '/:/installroot',
        '/sys:/installroot/sys',
        '/dev:/installroot/dev',
        '/proc:/installroot/proc',
        '/run/udev:/installroot/run/udev'
    ]
    if os.path.ismount('/boot'):
        bind_mounts.append('/boot:/installroot/boot')
    with _prepare_transaction(used_repos=used_repos,
                              target_userspace_info=target_userspace_info,
                              binds=bind_mounts
                              ) as (context, target_repoids, _unused):
        _transaction(context=context, stage='upgrade', target_repoids=target_repoids, tasks=tasks)


def perform_transaction_check(target_userspace_info, used_repos, tasks, xfs_present):
    """
    Perform DNF transaction check using our plugin
    """
    with _prepare_transaction(used_repos=used_repos,
                              target_userspace_info=target_userspace_info
                              ) as (context, target_repoids, userspace_info):
        with overlaygen.create_source_overlay(mounts_dir=userspace_info.mounts, scratch_dir=userspace_info.scratch,
                                              xfs_present=xfs_present) as overlay:
            utils.apply_yum_workaround(overlay.nspawn())
            with mounting.BindMount(source=overlay.target, target=os.path.join(context.base_dir, 'installroot')):
                _transaction(context=context, stage='check', target_repoids=target_repoids, tasks=tasks)


def perform_rpm_download(target_userspace_info, used_repos, tasks, xfs_present):
    """
    Perform RPM download including the transaction test using dnf with our plugin
    """
    with _prepare_transaction(used_repos=used_repos,
                              target_userspace_info=target_userspace_info
                              ) as (context, target_repoids, userspace_info):
        with overlaygen.create_source_overlay(mounts_dir=userspace_info.mounts, scratch_dir=userspace_info.scratch,
                                              xfs_present=xfs_present) as overlay:
            utils.apply_yum_workaround(overlay.nspawn())
            with mounting.BindMount(source=overlay.target, target=os.path.join(context.base_dir, 'installroot')):
                _transaction(context=context, stage='download', target_repoids=target_repoids, tasks=tasks, test=True)



import os

import pytest

from leapp.libraries.actor import library
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import produce_mocked, report_generic_mocked
from leapp.models import OSRelease


def _clean_leapp_envs(monkeypatch):
    """
    Clean all LEAPP environment variables before running the test to have
    fresh env.
    """
    for k, _ in os.environ.items():
        if k.startswith('LEAPP'):
            monkeypatch.delenv(k)


def test_leapp_env_vars(monkeypatch):
    _clean_leapp_envs(monkeypatch)
    monkeypatch.setenv('LEAPP_WHATEVER', '0')
    monkeypatch.setenv('LEAPP_VERBOSE', '1')
    monkeypatch.setenv('LEAPP_DEBUG', '1')
    monkeypatch.setenv('LEAPP_CURRENT_PHASE', 'test')
    monkeypatch.setenv('LEAPP_CURRENT_ACTOR', 'test')
    monkeypatch.setenv('TEST', 'test')
    monkeypatch.setenv('TEST2', 'test')

    assert len(library.get_env_vars()) == 1


def test_get_os_release_info(monkeypatch):
    monkeypatch.setattr('leapp.libraries.stdlib.api.produce', produce_mocked())
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())

    expected = OSRelease(
        release_id='rhel',
        name='Red Hat Enterprise Linux Server',
        pretty_name='Red Hat Enterprise Linux',
        version='7.6 (Maipo)',
        version_id='7.6',
        variant='Server',
        variant_id='server')
    assert expected == library.get_os_release('tests/files/os-release')

    assert not library.get_os_release('tests/files/non-existent-file')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']



from leapp.libraries import stdlib
from leapp.libraries.actor import kernelcmdlineconfig
from leapp.libraries.stdlib import api
from leapp.models import InstalledTargetKernelVersion, KernelCmdlineArg

KERNEL_VERSION = '1.2.3-4.x86_64.el8'


class MockedRun(object):
    def __init__(self):
        self.commands = []

    def __call__(self, cmd, *args, **kwargs):
        self.commands.append(cmd)
        return {}


def mocked_consume(*models):
    if InstalledTargetKernelVersion in models:
        return iter((InstalledTargetKernelVersion(version=KERNEL_VERSION),))
    return iter((
        KernelCmdlineArg(key='some_key1', value='some_value1'),
        KernelCmdlineArg(key='some_key2', value='some_value2')
    ))


def mocked_consume_no_args(*models):
    if InstalledTargetKernelVersion in models:
        return iter((InstalledTargetKernelVersion(version=KERNEL_VERSION),))
    return iter(())


def mocked_consume_no_version(*models):
    if InstalledTargetKernelVersion in models:
        return iter(())
    assert False and 'this should not be called'
    return iter(())


def test_kernelcmdline_config(monkeypatch):
    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'consume', mocked_consume)
    kernelcmdlineconfig.process()
    assert mocked_run.commands and len(mocked_run.commands) == 2
    assert ['grubby', '--update-kernel=/boot/vmlinuz-{}'.format(
        KERNEL_VERSION), '--args=some_key2=some_value2'] == mocked_run.commands.pop()
    assert ['grubby', '--update-kernel=/boot/vmlinuz-{}'.format(
        KERNEL_VERSION), '--args=some_key1=some_value1'] == mocked_run.commands.pop()


def test_kernelcmdline_config_no_args(monkeypatch):
    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'consume', mocked_consume_no_args)
    kernelcmdlineconfig.process()
    assert not mocked_run.commands


def test_kernelcmdline_config_no_version(monkeypatch):
    mocked_run = MockedRun()
    monkeypatch.setattr(stdlib, 'run', mocked_run)
    monkeypatch.setattr(api, 'consume', mocked_consume_no_version)
    kernelcmdlineconfig.process()
    assert not mocked_run.commands



from leapp.actors import Actor
from leapp.models import RepositoriesBlacklisted, RepositoriesFacts
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class RepositoriesBlacklist(Actor):
    """
    Generate list of Repositories ID that should be ignored by Leapp during upgrade process
    """

    name = 'repositories_blacklist'
    consumes = (RepositoriesFacts,)
    produces = (RepositoriesBlacklisted,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def _is_repo_enabled(self, repoid):
        for repos in self.consume(RepositoriesFacts):
            for repo_file in repos.repositories:
                for repo in repo_file.data:
                    if repo.repoid == repoid and repo.enabled:
                        return True
        return False

    def process(self):
        # blacklist crb repo <=> optional repo is not enabled
        if not self._is_repo_enabled('rhel-7-server-optional-rpms'):
            self.log.info("The optional repository is not enabled. Add the CRB repository to the blacklist.")
            self.produce(RepositoriesBlacklisted(
                repoids=['codeready-builder-for-rhel-8-x86_64-rpms']))



from leapp.models import Model, fields
from leapp.topics import SystemInfoTopic


class InstalledTargetKernelVersion(Model):
    """
    This message is used to propagate the version of the kernel that has been installed during the upgrade process.

    The version value is mainly used for boot loader entry manipulations, to know which boot entry to manipulate.
    """
    topic = SystemInfoTopic
    version = fields.String()



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, run
from leapp.models import StorageInfo, XFSPresence


def check_xfs_fstab(data):
    mountpoints = set()
    for entry in data:
        if entry.fs_vfstype == "xfs":
            mountpoints.add(entry.fs_file)

    return mountpoints


def check_xfs_mount(data):
    mountpoints = set()
    for entry in data:
        if entry.tp == "xfs":
            mountpoints.add(entry.mount)

    return mountpoints


def check_xfs_systemdmount(data):
    mountpoints = set()
    for entry in data:
        if entry.fs_type == "xfs":
            mountpoints.add(entry.path)

    return mountpoints


def is_xfs_without_ftype(mp):
    for l in run(['/usr/sbin/xfs_info', '{}'.format(mp)], split=True)['stdout']:
        if 'ftype=0' in l:
            return True

    return False


def check_xfs():
    storage_info_msgs = api.consume(StorageInfo)
    storage_info = next(storage_info_msgs, None)

    if list(storage_info_msgs):
        api.current_logger().warning('Unexpectedly received more than one StorageInfo message.')
    if not storage_info:
        raise StopActorExecutionError('Could not check if XFS is in use.',
                                      details={'details': 'Did not receive a StorageInfo message'})

    fstab_data = check_xfs_fstab(storage_info.fstab)
    mount_data = check_xfs_mount(storage_info.mount)
    systemdmount_data = check_xfs_systemdmount(storage_info.systemdmount)

    mountpoints = fstab_data | mount_data | systemdmount_data

    xfs_presence = XFSPresence()
    # By now, we only care for XFS without ftype in use for /var
    has_xfs_without_ftype = False
    for mp in ('/var', '/'):
        if mp in mountpoints:
            xfs_presence.present = True
            if is_xfs_without_ftype(mp):
                has_xfs_without_ftype = True

    xfs_presence.without_ftype = has_xfs_without_ftype
    api.produce(xfs_presence)



from leapp.actors import Actor
from leapp.libraries.actor.library import scan_repositories
from leapp.models import RepositoriesMap
from leapp.reporting import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class RepositoriesMapping(Actor):
    """
    Produces message containing repository mapping based on provided file.
    """

    name = 'repository_mapping'
    consumes = ()
    produces = (RepositoriesMap, Report,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        scan_repositories('/etc/leapp/files/repomap.csv')



from leapp.models import CustomTargetRepository, TargetRepositories, RepositoryData, \
    RepositoryFile, RepositoriesFacts, RepositoryMap, RepositoriesMap, RepositoriesSetupTasks, \
    RepositoriesBlacklisted


def test_minimal_execution(current_actor_context):
    current_actor_context.run()


def test_custom_repos(current_actor_context):
    custom = CustomTargetRepository(repoid='rhel-8-server-rpms',
                                    name='RHEL 8 Server (RPMs)',
                                    baseurl='https://.../dist/rhel/server/8/os',
                                    enabled=True)

    blacklisted = CustomTargetRepository(repoid='rhel-8-blacklisted-rpms',
                                         name='RHEL 8 Blacklisted (RPMs)',
                                         baseurl='https://.../dist/rhel/blacklisted/8/os',
                                         enabled=True)

    repos_blacklisted = RepositoriesBlacklisted(repoids=['rhel-8-blacklisted-rpms'])

    current_actor_context.feed(custom)
    current_actor_context.feed(blacklisted)
    current_actor_context.feed(repos_blacklisted)
    current_actor_context.run()

    assert current_actor_context.consume(TargetRepositories)

    custom_repos = current_actor_context.consume(TargetRepositories)[0].custom_repos
    assert len(custom_repos) == 1
    assert custom_repos[0].repoid == 'rhel-8-server-rpms'


def test_repositories_setup_tasks(current_actor_context):
    repositories_setup_tasks = RepositoriesSetupTasks(to_enable=['rhel-8-server-rpms',
                                                                 'rhel-8-blacklisted-rpms'])

    repos_blacklisted = RepositoriesBlacklisted(repoids=['rhel-8-blacklisted-rpms'])

    current_actor_context.feed(repositories_setup_tasks)
    current_actor_context.feed(repos_blacklisted)
    current_actor_context.run()
    assert current_actor_context.consume(TargetRepositories)

    rhel_repos = current_actor_context.consume(TargetRepositories)[0].rhel_repos
    assert len(rhel_repos) == 1
    assert rhel_repos[0].repoid == 'rhel-8-server-rpms'


def test_repos_mapping(current_actor_context):
    repos_data = [
        RepositoryData(repoid='rhel-7-server-rpms', name='RHEL 7 Server'),
        RepositoryData(repoid='rhel-7-blacklisted-rpms', name='RHEL 7 Blacklisted')]
    repos_files = [RepositoryFile(file='/etc/yum.repos.d/redhat.repo', data=repos_data)]
    facts = RepositoriesFacts(repositories=repos_files)

    mapping = [RepositoryMap(from_id='rhel-7-server-rpms',
                             to_id='rhel-8-for-x86_64-baseos-htb-rpms',
                             from_minor_version='all',
                             to_minor_version='all',
                             arch='x86_64',
                             repo_type='rpm'),
               RepositoryMap(from_id='rhel-7-server-rpms',
                             to_id='rhel-8-for-x86_64-appstream-htb-rpms',
                             from_minor_version='all',
                             to_minor_version='all',
                             arch='x86_64',
                             repo_type='rpm'),
               RepositoryMap(from_id='rhel-7-blacklist-rpms',
                             to_id='rhel-8-blacklist-rpms',
                             from_minor_version='all',
                             to_minor_version='all',
                             arch='x86_64',
                             repo_type='rpm')]
    repos_map = RepositoriesMap(repositories=mapping)

    repos_blacklisted = RepositoriesBlacklisted(repoids=['rhel-8-blacklisted-rpms'])

    current_actor_context.feed(facts)
    current_actor_context.feed(repos_map)
    current_actor_context.feed(repos_blacklisted)
    current_actor_context.run()
    assert current_actor_context.consume(TargetRepositories)

    rhel_repos = current_actor_context.consume(TargetRepositories)[0].rhel_repos
    assert len(rhel_repos) == 2
    assert set([repo.repoid for repo in rhel_repos]) == set(['rhel-8-for-x86_64-baseos-htb-rpms',
                                                             'rhel-8-for-x86_64-appstream-htb-rpms'])



import os

from leapp.exceptions import StopActorExecution, StopActorExecutionError
from leapp.libraries.common import reporting
from leapp.libraries.stdlib import api
from leapp.models import OSReleaseFacts


def skip_check():
    """ Check if an environment variable was used to skip this actor """
    if os.getenv('LEAPP_SKIP_CHECK_OS_RELEASE'):
        reporting.report_generic(
            title='Skipped OS release check',
            severity='high',
            summary='Source RHEL release check skipped via LEAPP_SKIP_CHECK_OS_RELEASE env var.',
        )
        return True
    return False


def check_os_version(supported_version):
    """ Check OS version and inhibit upgrade if not the same as supported ones """
    if not isinstance(supported_version, dict):
        api.current_logger().warning('The supported version value is invalid.')
        raise StopActorExecution()

    facts_messages = api.consume(OSReleaseFacts)
    facts = next(facts_messages, None)
    if list(facts_messages):
        api.current_logger().warning('Unexpectedly received more than one OSReleaseFacts message.')
    if not facts:
        raise StopActorExecutionError(
            'Could not check OS version', details={'details': 'No OSReleaseFacts facts found.'}
        )

    if facts.release_id not in supported_version:
        reporting.report_generic(
            title='Unsupported OS',
            summary='Only RHEL is supported by the upgrade process',
            flags=['inhibitor'],
        )
        return

    if not isinstance(supported_version[facts.release_id], list):
        raise StopActorExecutionError(
            'Invalid versions',
            details={'details': 'OS versions are invalid, please provide a valid list.'},
        )

    if facts.version_id not in supported_version[facts.release_id]:
        reporting.report_generic(
            title='Unsupported OS version',
            summary='The supported OS versions for the upgrade process: {}'.format(
                ', '.join(supported_version[facts.release_id])
            ),
            flags=['inhibitor'],
        )



import shutil

from leapp.actors import Actor
from leapp.libraries.common import dnfplugin
from leapp.libraries.stdlib import run
from leapp.models import FilteredRpmTransactionTasks, TargetUserSpaceInfo, TransactionCompleted, UsedTargetRepositories
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class DnfUpgradeTransaction(Actor):
    """
    Setup and call DNF upgrade command

    Based on previously calculated RPM transaction data, this actor will setup and call
    rhel-upgrade DNF plugin with necessary parameters
    """

    name = 'dnf_upgrade_transaction'
    consumes = (FilteredRpmTransactionTasks, TargetUserSpaceInfo, UsedTargetRepositories)
    produces = (TransactionCompleted,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        # FIXME: we hitting issue now because the network is down and rhsm
        # # is trying to connect to the server. Commenting this out for now
        # # so people will not be affected in case they do not have set a
        # # release and we will have time to fix it properly.
        # Make sure Subscription Manager OS Release is unset
        # cmd = ['subscription-manager', 'release', '--unset']
        # run(cmd)

        # FIXME: that's ugly hack, we should get info which file remove and
        # do it more nicely..
        run(['rm', '-f', '/etc/pki/product/69.pem'])

        used_repos = self.consume(UsedTargetRepositories)
        tasks = next(self.consume(FilteredRpmTransactionTasks), FilteredRpmTransactionTasks())
        target_userspace_info = next(self.consume(TargetUserSpaceInfo), None)

        dnfplugin.perform_transaction_install(tasks=tasks, used_repos=used_repos,
                                              target_userspace_info=target_userspace_info)
        self.produce(TransactionCompleted())
        userspace = next(self.consume(TargetUserSpaceInfo), None)
        if userspace:
            try:
                shutil.rmtree(userspace.path)
            except EnvironmentError:
                self.log.info("Failed to remove temporary userspace - error ignored", exc_info=True)



from leapp.actors import Actor
from leapp.libraries.actor.library import pes_events_scanner
from leapp.models import (InstalledRedHatSignedRPM, PESRpmTransactionTasks,
                          RepositoriesBlacklisted, RepositoriesSetupTasks,
                          RpmTransactionTasks)
from leapp.reporting import Report
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class PesEventsScanner(Actor):
    """
    Provides data about packages events from Package Evolution Service.

    After collecting data from a provided JSON file containing Package Evolution Service events, a
    message with relevant data will be produced to help DNF Upgrade transaction calculation.
    """

    name = 'pes_events_scanner'
    consumes = (InstalledRedHatSignedRPM, RpmTransactionTasks, RepositoriesBlacklisted)
    produces = (PESRpmTransactionTasks, RepositoriesSetupTasks, Report,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        pes_events_scanner('/etc/leapp/files/pes-events.json')



from leapp.actors import Actor
from leapp.libraries.actor.library import check_os_version, skip_check
from leapp.models import OSReleaseFacts
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckOSRelease(Actor):
    """
    Check if a supported release version of system's OS is in use. If not, inhibit upgrade process.

    Based on OS release collected facts, this actor will compare current release with supported
    versions. If a problem is found an inhibition message will be generated. This check can be
    skipped by using LEAPP_SKIP_CHECK_OS_RELEASE environment variable.
    """

    name = 'check_os_release'
    consumes = (OSReleaseFacts,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        supported_version = {'rhel': ['7.6']}
        if not skip_check():
            check_os_version(supported_version)



from leapp.models import ActiveKernelModulesFacts, ActiveKernelModule, KernelModuleParameter, WhitelistedKernelModules
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context


def test_actor_without_missing_drivers(current_actor_context):
    """ Tests CheckKernelDrivers actor by feeding it kernel drivers that
    are available on the RHEL8 system. Actor should NOT produce any report.
    """
    modules = [ActiveKernelModule(filename="foobar", parameters=[]),
               ActiveKernelModule(filename="barfoo", parameters=[])]
    current_actor_context.feed(ActiveKernelModulesFacts(kernel_modules=modules))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)


def test_actor_with_whitelisted_drivers(current_actor_context):
    """ Tests CheckKernelDrivers actor by feeding it kernel drivers that
    are whitelisted in the WhitelistKernelModule actor. Actor should NOT produce
    any report.
    """
    modules = WhitelistedKernelModules(whitelisted_modules=["cryptd", "floppy"])
    current_actor_context.feed(modules)
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



from leapp.libraries.stdlib import api, run
from leapp.models import SCTPConfig


def enable_sctp():
    """
    Enables the SCTP module by removing it from being black listed.
    """

    api.current_logger().info('Enabling SCTP.')
    run(['/usr/bin/sed', '-i', r's/^\s*blacklist.*sctp/#&/',
         '/etc/modprobe.d/sctp_diag-blacklist.conf',
         '/etc/modprobe.d/sctp-blacklist.conf'])
    api.current_logger().info('Enabled SCTP.')


def perform_update():
    for sctpconfig in api.consume(SCTPConfig):
        api.current_logger().info('Consuming sctp={}'.format(sctpconfig.wanted))
        if sctpconfig.wanted:
            enable_sctp()
            break



from leapp.actors import Actor
from leapp.libraries.common.reporting import report_with_remediation
from leapp.models import ActiveKernelModulesFacts, WhitelistedKernelModules, UdevAdmInfoData
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckKernelDrivers(Actor):
    """
    Actor checks if any loaded RHEL7 kernel driver is missing in the RHEL8.
    If yes, the upgrade process will be inhibited.

    Inhibition is done because missing kernel driver on the RHEL8 system may
    mean that the hardware using such driver would not work on the RHEL8.

    Note:
     - List of kernel drivers missing on the RHEL8 system is located in the
        files/removed_drivers.txt file.
     - Whitelisted modules that are not going to be reported are consumed from
        the WhitelistedKernelModules
    """
    name = 'check_kernel_drivers'
    consumes = (ActiveKernelModulesFacts, WhitelistedKernelModules, UdevAdmInfoData)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        with open('files/removed_drivers.txt', 'r') as removed:
            removed_drivers = []
            whitelisted_modules = set()
            collected_drivers = set()
            drivers_to_report = set()

            # Extracting kernel drivers from the files/removed_drivers.txt.
            for line in removed.readlines():
                token = line.strip()
                if token.startswith('#') or not token:
                    # We do not want comments or empty lines.
                    continue
                removed_drivers.append(token)

            # Consuming whitelisted kernel modules.
            for fact in self.consume(WhitelistedKernelModules):
                whitelisted_modules.update(fact.whitelisted_modules)

            # Collecting only non-whitelisted drivers that are part of the
            # files/removed_drivers.txt.
            for fact in self.consume(ActiveKernelModulesFacts):
                for active_module in fact.kernel_modules:
                    if active_module.filename in whitelisted_modules:
                        continue
                    if active_module.filename in removed_drivers:
                        collected_drivers.add(active_module.filename)

            # Going over the collected drivers and considering for reporting only
            # those drivers that are currently used by some device.
            udevadm_db = ''
            for fact in self.consume(UdevAdmInfoData):
                udevadm_db += fact.db
            for line in udevadm_db.split('\n'):
                if 'E: DRIVER=' in line:
                    _, driver = line.split('=')
                    if driver in collected_drivers:
                        drivers_to_report.add(driver)

            # In the end, we are only going to report drivers that are:
            # - removed in the RHEL8 (are part of files/removed_drivers.txt)
            # - not whitelisted
            # - currently being used by some device
            if drivers_to_report:
                title = ('Detected loaded kernel drivers which have been removed '
                         'in RHEL 8. Upgrade cannot proceed.')
                URL = ('https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/8/html-single/'
                       'considerations_in_adopting_rhel_8/index#removed-device-drivers_hardware-enablement')
                summary = ('Support for the following currently loaded RHEL 7 '
                           'device drivers has been removed in RHEL 8: \n     - {}'
                           '\nPlease see {} for details.'.format('\n     - '.join(drivers_to_report), URL))
                remediation = ('Please disable detected kernel drivers in '
                               'order to proceed with the upgrade process using the rmmod tool.')
                report_with_remediation(title=title,
                                        summary=summary,
                                        remediation=remediation,
                                        severity='high',
                                        flags=['inhibitor'])



from leapp.actors import Actor
from leapp.libraries.actor import scankernel
from leapp.models import InstalledTargetKernelVersion, TransactionCompleted
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class ScanInstalledTargetKernelVersion(Actor):
    """
    Scan for the version of the newly installed kernel

    This actor will query rpm for all kernel packages and reports the first matching el8 kernel RPM version.
    """

    name = 'scan_installed_target_kernel_version'
    consumes = (TransactionCompleted,)
    produces = (InstalledTargetKernelVersion,)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        scankernel.process()



from leapp.libraries.actor.library import get_os_release_info
from leapp.libraries.common import reporting
from leapp.libraries.common.testutils import produce_mocked, report_generic_mocked
from leapp.models import OSReleaseFacts


def test_get_os_release_info(monkeypatch):
    monkeypatch.setattr('leapp.libraries.stdlib.api.produce', produce_mocked())
    monkeypatch.setattr(reporting, 'report_generic', report_generic_mocked())

    expected = OSReleaseFacts(
        release_id='rhel',
        name='Red Hat Enterprise Linux Server',
        pretty_name='Red Hat Enterprise Linux',
        version='7.6 (Maipo)',
        version_id='7.6',
        variant='Server',
        variant_id='server')
    assert expected == get_os_release_info('tests/files/os-release')

    assert not get_os_release_info('tests/files/unexistent-file')
    assert reporting.report_generic.called == 1
    assert 'inhibitor' in reporting.report_generic.report_fields['flags']



from leapp.actors import Actor
from leapp.libraries.actor import setrelease
from leapp.models import TargetRHSMInfo
from leapp.tags import FirstBootPhaseTag, IPUWorkflowTag


class SetRhsmRelease(Actor):
    """
    No documentation has been provided for the set_rhsm_release actor.
    """

    name = 'set_rhsm_release'
    consumes = (TargetRHSMInfo,)
    produces = ()
    tags = (IPUWorkflowTag, FirstBootPhaseTag)

    def process(self):
        setrelease.process()



from leapp.libraries import stdlib
from leapp.models import InstalledRPM


def get_installed_rpms():
    rpm_cmd = [
        '/bin/rpm',
        '-qa',
        '--queryformat',
        r'%{NAME}|%{VERSION}|%{RELEASE}|%|EPOCH?{%{EPOCH}}:{(none)}||%|PACKAGER?{%{PACKAGER}}:{(none)}||%|'
        r'ARCH?{%{ARCH}}:{}||%|DSAHEADER?{%{DSAHEADER:pgpsig}}:{%|RSAHEADER?{%{RSAHEADER:pgpsig}}:{(none)}|}|\n'
    ]
    try:
        return stdlib.run(rpm_cmd, split=True)['stdout']
    except stdlib.CalledProcessError as err:
        error = 'Execution of {CMD} returned {RC}. Unable to find installed packages.'.format(CMD=err.command,
                                                                                              RC=err.exit_code)
        stdlib.api.current_logger().error(error)
        return []


def create_lookup(model, field, key, context=stdlib.api):
    """
    Create a lookup set from one of the model fields.

    :param model: model class
    :param field: model field, its value will be taken for lookup data
    :param key: property of the field's data that will be used to build a resulting set
    :param context: context of the execution
    """
    data = getattr(next((m for m in context.consume(model)), model()), field)
    try:
        return {getattr(obj, key) for obj in data} if data else {}
    except TypeError:
        # data is not iterable, not lookup can be built
        stdlib.api.current_logger().error(
                "{model}.{field}.{key} is not iterable, can't build lookup".format(
                    model=model, field=field, key=key))
        return {}


def has_package(model, package_name, context=stdlib.api):
    """
    Expects a model InstalledRedHatSignedRPM or InstalledUnsignedRPM.
    Can be useful in cases like a quick item presence check, ex. check in actor that
    a certain package is installed.

    :param model: model class
    :param package_name: package to be checked
    """
    if not (isinstance(model, type) and issubclass(model, InstalledRPM)):
        return False
    return package_name in create_lookup(model, field='items', key='name', context=context)



from leapp.models import Model, fields
from leapp.topics import TransactionTopic


class RpmTransactionTasks(Model):
    topic = TransactionTopic

    local_rpms = fields.List(fields.String(), default=[])
    to_install = fields.List(fields.String(), default=[])
    to_keep = fields.List(fields.String(), default=[])
    to_remove = fields.List(fields.String(), default=[])


class FilteredRpmTransactionTasks(RpmTransactionTasks):
    pass


class PESRpmTransactionTasks(RpmTransactionTasks):
    # Introduced because the framework struggles with solving dependency order of actors:
    # https://github.com/oamg/leapp/issues/491
    pass



from leapp.models import Model, fields
from leapp.topics import TransactionTopic


class TargetRepositoryBase(Model):
    topic = TransactionTopic
    repoid = fields.String()


class UsedTargetRepository(TargetRepositoryBase):
    pass


class RHELTargetRepository(TargetRepositoryBase):
    pass


class CustomTargetRepository(TargetRepositoryBase):
    name = fields.String()
    baseurl = fields.String()
    enabled = fields.Boolean(default=True)


class TargetRepositories(Model):
    topic = TransactionTopic
    rhel_repos = fields.List(fields.Model(RHELTargetRepository))
    custom_repos = fields.List(fields.Model(CustomTargetRepository), default=[])


class UsedTargetRepositories(Model):
    topic = TransactionTopic
    repos = fields.List(fields.Model(UsedTargetRepository))



from leapp.actors import Actor
from leapp.libraries.actor import initramgen
from leapp.models import (BootContent, RequiredUpgradeInitramPackages, TargetUserSpaceInfo, UpgradeDracutModule,
                          UsedTargetRepositories)
from leapp.tags import InterimPreparationPhaseTag, IPUWorkflowTag


class InitramDiskGenerator(Actor):
    """
    Creates the upgrade initram disk

    Creates an initram disk within a systemd-nspawn container using the target system userspace, including new kernel.
    The creation of the initram disk can be influenced with RequiredUpgradeInitramPackages and UpgradeDracutModule,
    which allow to specify additional packages to install in the target userspace and dracut modules to be included
    during the dracut execution.
    """

    name = 'initram_disk_generator'
    consumes = (RequiredUpgradeInitramPackages, TargetUserSpaceInfo, UpgradeDracutModule, UsedTargetRepositories)
    produces = (BootContent,)
    tags = (IPUWorkflowTag, InterimPreparationPhaseTag)

    def process(self):
        initramgen.process()



from leapp.actors import Actor
from leapp.libraries.actor.library import get_package_repository_data
from leapp.libraries.common.rpms import get_installed_rpms
from leapp.models import InstalledRPM, RPM
from leapp.tags import IPUWorkflowTag, FactsPhaseTag


class RpmScanner(Actor):
    """
    Provides data about installed RPM Packages.

    After collecting data from RPM query, a message with relevant data will be produced.
    """

    name = 'rpm_scanner'
    consumes = ()
    produces = (InstalledRPM,)
    tags = (IPUWorkflowTag, FactsPhaseTag)

    def process(self):
        output = get_installed_rpms()
        pkg_repos = get_package_repository_data()

        result = InstalledRPM()
        for entry in output:
            entry = entry.strip()
            if not entry:
                continue
            name, version, release, epoch, packager, arch, pgpsig = entry.split('|')
            repository = pkg_repos.get(name, '')
            result.items.append(RPM(
                name=name,
                version=version,
                epoch=epoch,
                packager=packager,
                arch=arch,
                release=release,
                pgpsig=pgpsig,
                repository=repository))
        self.produce(result)
